#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import contextlib
import copy
import datetime
import json
import logging
import os
from typing import Dict, List, Optional, Union

import oci.data_science
from ads.common.utils import batch_convert_case, get_value, snake_to_camel
from ads.config import COMPARTMENT_OCID, OCI_REGION_METADATA, PROJECT_OCID
from ads.jobs.builders.base import Builder
from ads.model.datascience_model import DataScienceModel
from ads.model.service.oci_datascience_model_version_set import (
    DataScienceModelVersionSet,
    ModelVersionSetNotExists,
    ModelVersionSetNotSaved,
)
from oci.data_science.models import UpdateModelDetails

logger = logging.getLogger(__name__)

_MVS_ID_ENV_VAR = "MODEL_VERSION_SET_ID"
_MVS_NAME_ENV_VAR = "MODEL_VERSION_SET_NAME"
_MVS_COMPARTMENT_ENV_VAR = "MODEL_VERSION_SET_COMPARTMENT_ID"
_MVS_URL = (
    "https://console.{region}.oraclecloud.com/data-science/model-version-sets/{id}"
)


class ModelVersionSet(Builder):
    """Represents Model Version Set.

    Attributes
    ----------
    id: str
        Model version set OCID.
    project_id: str
        Project OCID.
    compartment_id: str
        Compartment OCID.
    name: str
        Model version set name.
    description: str
        Model version set description.
    freeform_tags: Dict[str, str]
        Model version set freeform tags.
    defined_tags: Dict[str, Dict[str, object]]
        Model version set defined tags.
    details_link: str
        Link to details page in OCI console.

    Methods
    -------
    create(self, **kwargs) -> "ModelVersionSet"
        Creates a model version set.
    update(self, **kwargs) -> "ModelVersionSet"
        Updates a model version set.
    delete(self, delete_model: Optional[bool] = False) -> "ModelVersionSet":
        Removes a model version set.
    to_dict(self) -> dict
        Serializes model version set to a dictionary.
    from_id(cls, id: str) -> "ModelVersionSet"
        Gets an existing model version set by OCID.
    from_ocid(cls, ocid: str) -> "ModelVersionSet"
        Gets an existing model version set by OCID.
    from_name(cls, name: str) -> "ModelVersionSet"
        Gets an existing model version set by name.
    from_dict(cls, config: dict) -> "ModelVersionSet"
        Load a model version set instance from a dictionary of configurations.

    Examples
    --------
    >>> mvs = (ModelVersionSet()
    ...    .with_compartment_id(os.environ["PROJECT_COMPARTMENT_OCID"])
    ...    .with_project_id(os.environ["PROJECT_OCID"])
    ...    .with_name("test_experiment")
    ...    .with_description("Experiment number one"))
    >>> mvs.create()
    >>> mvs.model_add(model_ocid, version_label="Version label 1")
    >>> mvs.model_list()
    >>> mvs.details_link
    ... https://console.<region>.oraclecloud.com/data-science/model-version-sets/<ocid>
    >>> mvs.delete()
    """

    LIFECYCLE_STATE_ACTIVE = "ACTIVE"
    LIFECYCLE_STATE_DELETING = "DELETING"
    LIFECYCLE_STATE_DELETED = "DELETED"
    LIFECYCLE_STATE_FAILED = "FAILED"

    CONST_ID = "id"
    CONST_PROJECT_ID = "projectId"
    CONST_COMPARTMENT_ID = "compartmentId"
    CONST_NAME = "name"
    CONST_DESCRIPTION = "description"
    CONST_FREEFORM_TAG = "freeformTags"
    CONST_DEFINED_TAG = "definedTags"

    attribute_map = {
        CONST_ID: "id",
        CONST_PROJECT_ID: "project_id",
        CONST_COMPARTMENT_ID: "compartment_id",
        CONST_NAME: "name",
        CONST_DESCRIPTION: "description",
        CONST_FREEFORM_TAG: "freeform_tags",
        CONST_DEFINED_TAG: "defined_tags",
    }

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        """Initializes a model version set.

        Parameters
        ----------
        spec: (Dict, optional). Defaults to None.
            Object specification.

        kwargs: Dict
            Specification as keyword arguments.
            If 'spec' contains the same key as the one in kwargs,
            the value from kwargs will be used.

            - project_id: str
            - compartment_id: str
            - name: str
            - description: str
            - defined_tags: Dict[str, Dict[str, object]]
            - freeform_tags: Dict[str, str]
        """
        defaults = ModelVersionSet._load_default_properties()
        if spec:
            spec = {
                snake_to_camel(k): v
                for k, v in spec.items()
                if snake_to_camel(k) in self.attribute_map and v is not None
            }
            defaults.update(spec)

        spec = {
            snake_to_camel(k): v
            for k, v in kwargs.items()
            if snake_to_camel(k) in self.attribute_map and v is not None
        }
        defaults.update(spec)

        super().__init__(spec=defaults)
        self.dsc_model_version_set = DataScienceModelVersionSet(**self._spec)

    @property
    def project_id(self) -> str:
        return self.get_spec(self.CONST_PROJECT_ID)

    @project_id.setter
    def project_id(self, value: str):
        self.with_project_id(value)

    def with_project_id(self, project_id: str) -> "ModelVersionSet":
        """Sets the project OCID.

        Parameters
        ----------
        project_id: str
            The project OCID.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)
        """
        return self.set_spec(self.CONST_PROJECT_ID, project_id)

    @property
    def description(self) -> str:
        return self.get_spec(self.CONST_DESCRIPTION)

    @description.setter
    def description(self, value: str):
        self.with_description(value)

    def with_description(self, description: str) -> "ModelVersionSet":
        """Sets the description.

        Parameters
        ----------
        description: str
            The description of the model version set.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)

        """
        return self.set_spec(self.CONST_DESCRIPTION, description)

    @property
    def compartment_id(self) -> str:
        return self.get_spec(self.CONST_COMPARTMENT_ID)

    @compartment_id.setter
    def compartment_id(self, value: str):
        self.with_compartment_id(value)

    def with_compartment_id(self, compartment_id: str) -> "ModelVersionSet":
        """Sets the compartment OCID.

        Parameters
        ----------
        compartment_id: str
            The compartment OCID.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)
        """
        return self.set_spec(self.CONST_COMPARTMENT_ID, compartment_id)

    @property
    def name(self) -> str:
        return self.get_spec(self.CONST_NAME)

    @name.setter
    def name(self, value: str):
        self.with_name(value)

    def with_name(self, name: str) -> "ModelVersionSet":
        """Sets the name of the model version set.

        Parameters
        ----------
        name: str
            The name of the model version set.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)
        """
        return self.set_spec(self.CONST_NAME, name)

    @property
    def freeform_tags(self) -> Dict[str, str]:
        return self.get_spec(self.CONST_FREEFORM_TAG)

    @freeform_tags.setter
    def freeform_tags(self, value: Dict[str, str]):
        self.with_freeform_tags(**value)

    def with_freeform_tags(self, **kwargs: Dict[str, str]) -> "ModelVersionSet":
        """Sets freeform tags.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)
        """
        return self.set_spec(self.CONST_FREEFORM_TAG, kwargs)

    @property
    def defined_tags(self) -> Dict[str, Dict[str, object]]:
        return self.get_spec(self.CONST_FREEFORM_TAG)

    @defined_tags.setter
    def defined_tags(self, value: Dict[str, Dict[str, object]]):
        self.with_defined_tags(**value)

    def with_defined_tags(
        self, **kwargs: Dict[str, Dict[str, object]]
    ) -> "ModelVersionSet":
        """Sets defined tags.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)
        """
        return self.set_spec(self.CONST_DEFINED_TAG, kwargs)

    def create(self, **kwargs) -> "ModelVersionSet":
        """Creates a model version set.

        Parameters
        ----------
        kwargs
            Additional keyword arguments.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self)
        """
        if not self.name:
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M")
            self.name = f"model-version-set-{timestamp}"

        payload = copy.deepcopy(self._spec)
        payload.update({"name": self.name})

        if not payload.get(self.CONST_COMPARTMENT_ID, None):
            raise ValueError(
                "Compartment id is required. Specify compartment id via `with_compartment_id()`."
            )

        payload.pop("id", None)
        logger.debug(f"Creating a model version set with payload {payload}")
        self.dsc_model_version_set = DataScienceModelVersionSet(**payload).create()
        # self.with_id(self.dsc_model_version_set.id)
        self._update_from_dsc_model(self.dsc_model_version_set)
        return self

    def update(self) -> "ModelVersionSet":
        """Updates a model version set.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self).
        """
        payload = copy.deepcopy(self._spec)
        logger.debug(f"Updating a model version set with payload {payload}")
        self.dsc_model_version_set = DataScienceModelVersionSet(**payload).update()
        self._update_from_dsc_model(self.dsc_model_version_set)
        return self

    def delete(self, delete_model: Optional[bool] = False) -> "ModelVersionSet":
        """Removes a model version set.

        Parameters
        ----------
        delete_model: (bool, optional). Defaults to False.
            By default, this parameter is false. A model version set can only be
            deleted if all the models associate with it are already in the DELETED state.
            You can optionally specify the deleteRelatedModels boolean query parameters to
            true, which deletes all associated models for you.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self).
        """
        self.dsc_model_version_set.delete(delete_model)

    def model_add(
        self, model_id: str, version_label: Optional[str] = None, **kwargs
    ) -> None:
        """Adds new model to model version set.

        Parameters
        ----------
        model_id: str
            The OCID of the model which needs to be associated with the model version set.
        version_label: str
            The model version label.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ModelVersionSetNotSaved: If model version set has not been saved yet.
        """
        if not self.id:
            raise ModelVersionSetNotSaved("Model version set needs to be saved.")

        self.dsc_model_version_set.client.update_model(
            model_id,
            UpdateModelDetails(
                model_version_set_id=self.id, version_label=version_label
            ),
            **kwargs,
        )

    @classmethod
    def _load_default_properties(cls) -> Dict:
        """
        Load default properties from environment variables, notebook session, etc.

        Returns
        -------
        Dict
            A dictionary of default properties.
        """
        defaults = {}
        compartment_ocid = COMPARTMENT_OCID
        if compartment_ocid:
            defaults[cls.CONST_COMPARTMENT_ID] = compartment_ocid
        if PROJECT_OCID:
            defaults[cls.CONST_PROJECT_ID] = PROJECT_OCID

        timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M")
        defaults[cls.CONST_NAME] = f"model-version-set-{timestamp}"

        return defaults

    @property
    def id(self) -> Optional[str]:
        """The OCID of the model version set."""
        if self.dsc_model_version_set:
            return self.dsc_model_version_set.id
        return None

    @property
    def status(self) -> Optional[str]:
        """Status of the model version set.

        Returns
        -------
        str
            Status of the model version set.
        """
        if self.dsc_model_version_set:
            return self.dsc_model_version_set.status
        return None

    @property
    def details_link(self) -> str:
        """Link to details page in OCI console.

        Returns
        -------
        str
            Link to details page in OCI console.
        """
        if self.dsc_model_version_set:
            signer = self.dsc_model_version_set.auth or {}
            if "region" in signer.get("config", {}):
                region = signer["config"]["region"]
            else:
                region = json.loads(OCI_REGION_METADATA)["regionIdentifier"]
            return _MVS_URL.format(region=region, id=self.id)

        return None

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            "modelVersionSet"
        """
        return "modelVersionSet"

    @classmethod
    def list(cls, compartment_id: str = None, **kwargs) -> List["ModelVersionSet"]:
        """
        List model version sets in a given compartment.

        Parameters
        ----------
        compartment_id: str
            The OCID of compartment.
        kwargs
            Additional keyword arguments for filtering model version sets.

        Returns
        -------
        List[ModelVersionSet]
            The list of model version sets.
        """
        return [
            cls.from_dsc_model_version_set(model_version_set)
            for model_version_set in DataScienceModelVersionSet.list_resource(
                compartment_id, **kwargs
            )
        ]

    def models(self, **kwargs) -> List[DataScienceModel]:
        """Gets list of models associated with a model version set.

        Parameters
        ----------
        kwargs:
            project_id: str
                Project OCID.
           lifecycle_state: str
                Filter results by the specified lifecycle state. Must be a valid state for the resource type.
                Allowed values are: "ACTIVE", "DELETED", "FAILED", "INACTIVE"

            Can be any attribute that `oci.data_science.data_science_client.DataScienceClient.list_models.` accepts.

        Returns
        -------
        List[DataScienceModel]
            List of models associated with the model version set.

        Raises
        ------
        ModelVersionSetNotSaved: If model version set has not been saved yet.
        """
        if not self.id:
            raise ModelVersionSetNotSaved("Model version set needs to be saved.")

        return DataScienceModel.list(
            compartment_id=self.compartment_id,
            model_version_set_name=self.name,
            **kwargs,
        )

    @classmethod
    def from_dsc_model_version_set(
        cls, dsc_model_version_set: DataScienceModelVersionSet
    ) -> "ModelVersionSet":
        """Initialize a ModelVersionSet instance from a DataScienceModelVersionSet.

        Parameters
        ----------
        dsc_model_version_set: DataScienceModelVersionSet
            An instance of DataScienceModelVersionSet.

        Returns
        -------
        ModelVersionSet
            An instance of ModelVersionSet.
        """
        return cls()._update_from_dsc_model(dsc_model_version_set)

    def _update_from_dsc_model(
        self, dsc_model_version_set: oci.data_science.models.ModelVersionSet
    ) -> "ModelVersionSet":
        """Update the properties from an OCI data science model version set model.

        Parameters
        ----------
        dsc_model_version_set: oci.data_science.models.ModelVersionSet
            An OCI data science model version set model.

        Returns
        -------
        ModelVersionSet
            The ModelVersionSet instance (self).
        """
        self.dsc_model_version_set = dsc_model_version_set
        for infra_attr, dsc_attr in self.attribute_map.items():
            value = get_value(dsc_model_version_set, dsc_attr)
            if value:
                self._spec[infra_attr] = get_value(dsc_model_version_set, dsc_attr)
        return self

    @classmethod
    def from_id(cls, id: str) -> "ModelVersionSet":
        """Gets an existing model version set by OCID.

        Parameters
        ----------
        id: str
            The model version set OCID.

        Returns
        -------
        ModelVersionSet
            An instance of ModelVersionSet.
        """
        return cls.from_dsc_model_version_set(DataScienceModelVersionSet.from_ocid(id))

    @classmethod
    def from_ocid(cls, ocid: str) -> "ModelVersionSet":
        """Gets an existing model version set by OCID.

        Parameters
        ----------
        id: str
            The model version set OCID.

        Returns
        -------
        ModelVersionSet
            An instance of ModelVersionSet.
        """
        return cls.from_id(ocid)

    @classmethod
    def from_name(
        cls, name: str, compartment_id: Optional[str] = None
    ) -> "ModelVersionSet":
        """Gets an existing model version set by name.

        Parameters
        ----------
        name: str
            The model version set name.
        compartment_id: (str, optional). Defaults to None.
            Compartment OCID of the OCI resources. If `compartment_id` is not specified,
            the value will be taken from environment variables.

        Returns
        -------
        ModelVersionSet
            An instance of ModelVersionSet.
        """
        compartment_id = compartment_id or COMPARTMENT_OCID
        return cls.from_dsc_model_version_set(
            DataScienceModelVersionSet.from_name(
                name=name, compartment_id=compartment_id
            )
        )

    def to_dict(self) -> dict:
        """
        Serializes model version set to a dictionary.

        Returns
        -------
        dict
            The model version set serialized as a dictionary.
        """
        return {
            "kind": self.kind,
            "type": self.type,
            "spec": batch_convert_case(self._spec, "camel"),
        }

    @classmethod
    def from_dict(cls, config: dict) -> "ModelVersionSet":
        """
        Load a model version set instance from a dictionary of configurations.

        Parameters
        ----------
        config: dict
            A dictionary of configurations.

        Returns
        -------
        ModelVersionSet
            The model version set instance.
        """
        return cls(spec=batch_convert_case(config.get("spec"), "snake"))

    def __getattr__(self, item):
        if f"with_{item}" in self.__dir__():
            return self.get_spec(item)
        raise AttributeError(f"Attribute {item} not found.")


def _extract_model_version_set_id(
    model_version_set: Optional[Union[str, ModelVersionSet]] = None,
    compartment_id: Optional[str] = None,
) -> str:
    """Extracts model version set id.
    If `model_version_set` attribute not specified, then the `model_version_set_id`
    will be extracted from the environment variables.

    Parameters
    ----------
    model_version_set: (Union[str, ModelVersionSet], optional). Defaults to None.
        The model version set information. Can be name, ocid or `ModelVersionSet` instance.
        If `model_version_set` not provided then id will be extracted from environment variables.
    compartment_id: (str, optional). Defaults to value from the environment variables.
            The compartment OCID.

    Returns
    -------
    str
        The model version set OCID.
    """
    compartment_id = (
        compartment_id or os.environ.get(_MVS_COMPARTMENT_ENV_VAR) or COMPARTMENT_OCID
    )
    if not model_version_set:
        return os.environ.get(_MVS_ID_ENV_VAR)
    if isinstance(model_version_set, ModelVersionSet):
        return model_version_set.id
    if isinstance(model_version_set, str):
        if model_version_set.lower().startswith("ocid"):
            return model_version_set
        # search model version set by name
        return ModelVersionSet.from_name(
            name=model_version_set, compartment_id=compartment_id
        ).id
    return None


@contextlib.contextmanager
def experiment(name: str, create_if_not_exists: Optional[bool] = True, **kwargs: Dict):
    """Context manager helping to operate with model version set.

    Parameters
    ----------
    name: str
        The name of the model version set.
    create_if_not_exists: (bool, optional). Defaults to True.
        Creates model version set if not exists.

    kwargs: (Dict, optional).
        compartment_id: (str, optional). Defaults to value from the environment variables.
            The compartment OCID.
        project_id: (str, optional). Defaults to value from the environment variables.
            The project OCID.
        description: (str, optional). Defaults to None.
            The description of the model version set.

    Yields
    ------
    ModelVersionSet
        The model version set object.
    """
    compartment_id = kwargs.pop("compartment_id", None) or COMPARTMENT_OCID
    project_id = kwargs.pop("project_id", None) or PROJECT_OCID

    try:
        mvs = ModelVersionSet.from_name(name=name, compartment_id=compartment_id)
    except ModelVersionSetNotExists:
        if not create_if_not_exists:
            raise
        else:
            mvs = (
                ModelVersionSet()
                .with_compartment_id(compartment_id)
                .with_project_id(project_id)
                .with_name(name)
                .with_description(kwargs.pop("description", ""))
                .create()
            )

    try:
        os.environ[_MVS_ID_ENV_VAR] = mvs.id
        os.environ[_MVS_NAME_ENV_VAR] = mvs.name
        os.environ[_MVS_COMPARTMENT_ENV_VAR] = compartment_id

        yield mvs
    finally:
        os.environ.pop(_MVS_ID_ENV_VAR, None)
        os.environ.pop(_MVS_NAME_ENV_VAR, None)
        os.environ.pop(_MVS_COMPARTMENT_ENV_VAR, None)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/from typing import Dict


from typing import Dict, List
from ads.jobs.builders.base import Builder

MODEL_DEPLOYMENT_RUNTIME_KIND = "runtime"


class ModelDeploymentRuntimeType:
    CONDA = "conda"
    CONTAINER = "container"


class OCIModelDeploymentRuntimeType:
    CONDA = "DEFAULT"
    CONTAINER = "OCIR_CONTAINER"


class ModelDeploymentRuntime(Builder):
    """A class used to represent a Model Deployment Runtime.

    Attributes
    ----------
    env: Dict
        The environment variables of model deployment runtime.
    deployment_mode: str
        The deployment mode of model deployment.
    input_stream_ids: List
        The input stream ids of model deployment.
    output_stream_ids: List
        The output stream ids of model deployment.
    model_uri: str
        The model uri of model deployment.

    Methods
    -------
    with_env(env)
        Sets the environment variables of model deployment
    with_deployment_mode(deployment_mode)
        Sets the deployment mode of model deployment
    with_input_stream_ids(input_stream_ids)
        Sets the input stream ids of model deployment
    with_output_stream_ids(output_stream_ids)
        Sets the output stream ids of model deployment
    with_model_uri(model_uri)
        Sets the model uri of model deployment
    """

    CONST_MODEL_ID = "modelId"
    CONST_MODEL_URI = "modelUri"
    CONST_ENV = "env"
    CONST_ENVIRONMENT_VARIABLES = "environmentVariables"
    CONST_ENVIRONMENT_CONFIG_TYPE = "environmentConfigurationType"
    CONST_DEPLOYMENT_MODE = "deploymentMode"
    CONST_INPUT_STREAM_IDS = "inputStreamIds"
    CONST_OUTPUT_STREAM_IDS = "outputStreamIds"
    CONST_ENVIRONMENT_CONFIG_DETAILS = "environmentConfigurationDetails"

    attribute_map = {
        CONST_ENV: "env",
        CONST_ENVIRONMENT_VARIABLES: "environment_variables",
        CONST_INPUT_STREAM_IDS: "input_stream_ids",
        CONST_OUTPUT_STREAM_IDS: "output_stream_ids",
        CONST_DEPLOYMENT_MODE: "deployment_mode",
        CONST_MODEL_URI: "model_uri",
    }

    ENVIRONMENT_CONFIG_DETAILS_PATH = (
        "model_deployment_configuration_details.environment_configuration_details"
    )
    STREAM_CONFIG_DETAILS_PATH = (
        "model_deployment_configuration_details.stream_configuration_details"
    )
    MODEL_CONFIG_DETAILS_PATH = (
        "model_deployment_configuration_details.model_configuration_details"
    )

    payload_attribute_map = {
        CONST_ENV: f"{ENVIRONMENT_CONFIG_DETAILS_PATH}.environment_variables",
        CONST_INPUT_STREAM_IDS: f"{STREAM_CONFIG_DETAILS_PATH}.input_stream_ids",
        CONST_OUTPUT_STREAM_IDS: f"{STREAM_CONFIG_DETAILS_PATH}.output_stream_ids",
        CONST_DEPLOYMENT_MODE: "deployment_mode",
        CONST_MODEL_URI: f"{MODEL_CONFIG_DETAILS_PATH}.model_id",
    }

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        super().__init__(spec, **kwargs)

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            runtime
        """
        return MODEL_DEPLOYMENT_RUNTIME_KIND

    @property
    def env(self) -> Dict:
        """The environment variables of model deployment.

        Returns
        -------
        Dict
            The environment variables of model deployment.
        """
        return self.get_spec(self.CONST_ENV, {})

    def with_env(self, env: Dict) -> "ModelDeploymentRuntime":
        """Sets the environment variables of model deployment.

        Parameters
        ----------
        env: Dict
            The environment variables of model deployment.

        Returns
        -------
        ModelDeploymentRuntime
            The ModelDeploymentRuntime instance (self).
        """
        return self.set_spec(self.CONST_ENV, env)

    @property
    def deployment_mode(self) -> str:
        """The deployment mode of model deployment.

        Returns
        -------
        str
            The deployment mode of model deployment.
        """
        return self.get_spec(self.CONST_DEPLOYMENT_MODE, None)

    def with_deployment_mode(self, deployment_mode: str) -> "ModelDeploymentRuntime":
        """Sets the deployment mode of model deployment.
        Can be HTTPS_ONLY or STREAM_ONLY.

        Parameters
        ----------
        deployment_mode: str
            The deployment mode of model deployment.

        Returns
        -------
        ModelDeploymentRuntime
            The ModelDeploymentRuntime instance (self).
        """
        return self.set_spec(self.CONST_DEPLOYMENT_MODE, deployment_mode)

    @property
    def input_stream_ids(self) -> List[str]:
        """The input stream ids of model deployment.

        Returns
        -------
        List
            The input stream ids of model deployment.
        """
        return self.get_spec(self.CONST_INPUT_STREAM_IDS, [])

    def with_input_stream_ids(
        self, input_stream_ids: List[str]
    ) -> "ModelDeploymentRuntime":
        """Sets the input stream ids of model deployment.

        Parameters
        ----------
        input_stream_ids: List
            The input stream ids of model deployment.

        Returns
        -------
        ModelDeploymentRuntime
            The ModelDeploymentRuntime instance (self).
        """
        return self.set_spec(self.CONST_INPUT_STREAM_IDS, input_stream_ids)

    @property
    def output_stream_ids(self) -> List[str]:
        """The output stream ids of model deployment.

        Returns
        -------
        List
            The output stream ids of model deployment.
        """
        return self.get_spec(self.CONST_OUTPUT_STREAM_IDS, [])

    def with_output_stream_ids(
        self, output_stream_ids: List[str]
    ) -> "ModelDeploymentRuntime":
        """Sets the output stream ids of model deployment.

        Parameters
        ----------
        output_stream_ids: List
            The output stream ids of model deployment.

        Returns
        -------
        ModelDeploymentRuntime
            The ModelDeploymentRuntime instance (self).
        """
        return self.set_spec(self.CONST_OUTPUT_STREAM_IDS, output_stream_ids)

    @property
    def model_uri(self) -> str:
        """The model uri of model deployment.

        Returns
        -------
        List
            The model uri of model deployment.
        """
        return self.get_spec(self.CONST_MODEL_URI, None)

    def with_model_uri(self, model_uri: str) -> "ModelDeploymentRuntime":
        """Sets the model uri of model deployment.

        Parameters
        ----------
        model_uri: str
            The model uri of model deployment.

        Returns
        -------
        ModelDeploymentRuntime
            The ModelDeploymentRuntime instance (self).
        """
        return self.set_spec(self.CONST_MODEL_URI, model_uri)


class ModelDeploymentCondaRuntime(ModelDeploymentRuntime):
    """A class used to represent a Model Deployment Conda Runtime.

    Examples
    --------
    >>> conda_runtime = ModelDeploymentCondaRuntime()
    ...        .with_env({"key":"value"})
    ...        .with_deployment_mode("HTTPS_ONLY")
    ...        .with_model_uri(<model_uri>)
    >>> conda_runtime.to_dict()
    """

    @property
    def type(self) -> str:
        """The type of the object as showing in YAML.

        Returns
        -------
        str
            conda
        """
        return ModelDeploymentRuntimeType.CONDA

    @property
    def environment_config_type(self) -> str:
        """The environment config type of model deployment.

        Returns
        -------
        str
            DEFAULT
        """
        return OCIModelDeploymentRuntimeType.CONDA


class ModelDeploymentContainerRuntime(ModelDeploymentRuntime):
    """A class used to represent a Model Deployment Container Runtime.

    Attributes
    ----------
    image: str
        The image of model deployment container runtime.
    image_digest: str
        The image digest of model deployment container runtime.
    cmd: List
        The cmd of model deployment container runtime.
    entrypoint: List
        The entrypoint of model deployment container runtime.
    server_port: int
        The server port of model deployment container runtime.
    health_check_port: int
        The health check port of model deployment container runtime.

    Methods
    -------
    with_image(image)
        Sets the image of model deployment container runtime
    with_image_digest(image_digest)
        Sets the image digest of model deployment container runtime
    with_cmd(cmd)
        Sets the cmd of model deployment container runtime
    with_entrypoint(entrypoint)
        Sets the entrypoint of model deployment container runtime
    with_server_port(server_port)
        Sets the server port of model deployment container runtime
    with_health_check_port(health_check_port)
        Sets the health check port of model deployment container runtime

    Examples
    --------
    Build runtime from builder apis:
    >>> container_runtime = ModelDeploymentContainerRuntime()
    ...        .with_image(<image>)
    ...        .with_image_digest(<image_digest>)
    ...        .with_entrypoint(<entrypoint>)
    ...        .with_server_port(<server_port>)
    ...        .with_health_check_port(<health_check_port>)
    ...        .with_env({"key":"value"})
    ...        .with_deployment_mode("HTTPS_ONLY")
    ...        .with_model_uri(<model_uri>)
    >>> container_runtime.to_dict()

    Build runtime from yaml:
    >>> container_runtime = ModelDeploymentContainerRuntime.from_yaml(uri=<path_to_yaml>)
    """

    CONST_IMAGE = "image"
    CONST_IMAGE_DIGEST = "imageDigest"
    CONST_CMD = "cmd"
    CONST_ENTRYPOINT = "entrypoint"
    CONST_SERVER_PORT = "serverPort"
    CONST_HEALTH_CHECK_PORT = "healthCheckPort"

    attribute_map = {
        **ModelDeploymentRuntime.attribute_map,
        CONST_IMAGE: "image",
        CONST_IMAGE_DIGEST: "image_digest",
        CONST_CMD: "cmd",
        CONST_ENTRYPOINT: "entrypoint",
        CONST_SERVER_PORT: "server_port",
        CONST_HEALTH_CHECK_PORT: "health_check_port",
    }

    payload_attribute_map = {
        **ModelDeploymentRuntime.payload_attribute_map,
        CONST_IMAGE: f"{ModelDeploymentRuntime.ENVIRONMENT_CONFIG_DETAILS_PATH}.image",
        CONST_IMAGE_DIGEST: f"{ModelDeploymentRuntime.ENVIRONMENT_CONFIG_DETAILS_PATH}.image_digest",
        CONST_CMD: f"{ModelDeploymentRuntime.ENVIRONMENT_CONFIG_DETAILS_PATH}.cmd",
        CONST_ENTRYPOINT: f"{ModelDeploymentRuntime.ENVIRONMENT_CONFIG_DETAILS_PATH}.entrypoint",
        CONST_SERVER_PORT: f"{ModelDeploymentRuntime.ENVIRONMENT_CONFIG_DETAILS_PATH}.server_port",
        CONST_HEALTH_CHECK_PORT: f"{ModelDeploymentRuntime.ENVIRONMENT_CONFIG_DETAILS_PATH}.health_check_port",
    }

    @property
    def type(self) -> str:
        """The type of the object as showing in YAML.

        Returns
        -------
        str
            container
        """
        return ModelDeploymentRuntimeType.CONTAINER

    @property
    def environment_config_type(self) -> str:
        """The environment config type of model deployment.

        Returns
        -------
        str
            OCIR_CONTAINER
        """
        return OCIModelDeploymentRuntimeType.CONTAINER

    @property
    def image(self) -> str:
        """The image of model deployment container runtime.

        Returns
        -------
        str
            The image of model deployment container runtime.
        """
        return self.get_spec(self.CONST_IMAGE, None)

    def with_image(self, image: str) -> "ModelDeploymentContainerRuntime":
        """Sets the image of model deployment container runtime.

        Parameters
        ----------
        image: str
            The image of model deployment container runtime.

        Returns
        -------
        ModelDeploymentContainerRuntime
            The ModelDeploymentContainerRuntime instance (self).
        """
        return self.set_spec(self.CONST_IMAGE, image)

    @property
    def image_digest(self) -> str:
        """The image digest of model deployment container runtime.

        Returns
        -------
        str
            The image digest of model deployment container runtime.
        """
        return self.get_spec(self.CONST_IMAGE_DIGEST, None)

    def with_image_digest(self, image_digest: str) -> "ModelDeploymentContainerRuntime":
        """Sets the image digest of model deployment container runtime.

        Parameters
        ----------
        image_digest: str
            The image digest of model deployment container runtime.

        Returns
        -------
        ModelDeploymentContainerRuntime
            The ModelDeploymentContainerRuntime instance (self).
        """
        return self.set_spec(self.CONST_IMAGE_DIGEST, image_digest)

    @property
    def cmd(self) -> List[str]:
        """The command lines of model deployment container runtime.

        Returns
        -------
        List
            The command lines of model deployment container runtime.
        """
        return self.get_spec(self.CONST_CMD, [])

    def with_cmd(self, cmd: List[str]) -> "ModelDeploymentContainerRuntime":
        """Sets the cmd of model deployment container runtime.

        Parameters
        ----------
        cmd: List
            The cmd of model deployment container runtime.

        Returns
        -------
        ModelDeploymentContainerRuntime
            The ModelDeploymentContainerRuntime instance (self).
        """
        return self.set_spec(self.CONST_CMD, cmd)

    @property
    def entrypoint(self) -> List[str]:
        """The entry point of model deployment container runtime.

        Returns
        -------
        List
            The entry point of model deployment container runtime.
        """
        return self.get_spec(self.CONST_ENTRYPOINT, [])

    def with_entrypoint(
        self, entrypoint: List[str]
    ) -> "ModelDeploymentContainerRuntime":
        """Sets the entrypoint of model deployment container runtime.

        Parameters
        ----------
        entrypoint: List
            The entrypoint of model deployment container runtime.

        Returns
        -------
        ModelDeploymentContainerRuntime
            The ModelDeploymentContainerRuntime instance (self).
        """
        return self.set_spec(self.CONST_ENTRYPOINT, entrypoint)

    @property
    def server_port(self) -> int:
        """The server port of model deployment container runtime.

        Returns
        -------
        int
            The server port of model deployment container runtime.
        """
        return self.get_spec(self.CONST_SERVER_PORT, None)

    def with_server_port(self, server_port: int) -> "ModelDeploymentContainerRuntime":
        """Sets the server port of model deployment container runtime.

        Parameters
        ----------
        server_port: List
            The server port of model deployment container runtime.

        Returns
        -------
        ModelDeploymentContainerRuntime
            The ModelDeploymentContainerRuntime instance (self).
        """
        return self.set_spec(self.CONST_SERVER_PORT, server_port)

    @property
    def health_check_port(self) -> int:
        """The health check port of model deployment container runtime.

        Returns
        -------
        int
            The health check port of model deployment container runtime.
        """
        return self.get_spec(self.CONST_HEALTH_CHECK_PORT, None)

    def with_health_check_port(
        self, health_check_port: int
    ) -> "ModelDeploymentContainerRuntime":
        """Sets the health check port of model deployment container runtime.

        Parameters
        ----------
        health_check_port: List
            The health check port of model deployment container runtime.

        Returns
        -------
        ModelDeploymentContainerRuntime
            The ModelDeploymentContainerRuntime instance (self).
        """
        return self.set_spec(self.CONST_HEALTH_CHECK_PORT, health_check_port)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
from abc import ABC, abstractmethod
import dataclasses
from typing import Dict, Union, Optional, List
from enum import Enum

import fsspec
import yaml
from ads.common import logger

try:
    from yaml import CSafeDumper as dumper
    from yaml import CSafeLoader as loader
except:
    from yaml import SafeDumper as dumper
    from yaml import SafeLoader as loader


class SideEffect(Enum):
    CONVERT_KEYS_TO_LOWER = "lower"
    CONVERT_KEYS_TO_UPPER = "upper"


class Serializable(ABC):
    """Base class that represents a serializable item.

    Methods
    -------
    to_dict(self) -> dict
        Serializes the object into a dictionary.
    from_dict(cls, obj_dict) -> cls
        Returns an instance of the class instantiated from the dictionary provided.
    _write_to_file(s, uri, **kwargs)
        Write string s into location specified by uri
    _read_from_file(uri, **kwargs)
        Returns contents from location specified by URI
    to_json(self, uri=None, **kwargs)
        Returns object serialized as a JSON string
    from_json(cls, json_string=None, uri=None, **kwargs)
        Creates an object from JSON string provided or from URI location containing JSON string
    to_yaml(self, uri=None, **kwargs)
        Returns object serialized as a YAML string
    from_yaml(cls, yaml_string=None, uri=None, **kwargs)
        Creates an object from YAML string provided or from URI location containing YAML string
    from_string(cls, obj_string=None: str, uri=None, **kwargs)
        Creates an object from string provided or from URI location containing string
    """

    @abstractmethod
    def to_dict(self, **kwargs) -> dict:
        """Serializes instance of class into a dictionary.

        Returns
        -------
        Dict
            A dictionary.
        """
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, obj_dict: dict) -> "Serializable":
        """Returns an instance of the class instantiated by the dictionary provided.

        Parameters
        ----------
        obj_dict: (dict)
            Dictionary representation of the object.

        Returns
        -------
        Serializable
            A Serializable instance.
        """
        pass

    @staticmethod
    def _write_to_file(s: str, uri: str, **kwargs) -> None:
        """Write string s into location specified by uri.

        Parameters
        ----------
        s: (string)
            content
        uri: (string)
            URI location to save string s

        kwargs
        ------
        keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Returns
        -------
        None
            Nothing.
        """
        with fsspec.open(uri, "w", **kwargs) as f:
            f.write(s)

    @staticmethod
    def _read_from_file(uri: str, **kwargs) -> str:
        """Returns contents from location specified by URI

        Parameters
        ----------
        uri: (string)
            URI location

        kwargs
        ------
        keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Returns
        -------
            string: Contents in file specified by URI
        """
        with fsspec.open(uri, "r", **kwargs) as f:
            return f.read()

    def to_json(
        self, uri: str = None, encoder: callable = json.JSONEncoder, **kwargs
    ) -> str:
        """Returns object serialized as a JSON string

        Parameters
        ----------
        uri: (string, optional)
            URI location to save the JSON string. Defaults to None.
        encoder: (callable, optional)
            Encoder for custom data structures. Defaults to JSONEncoder.

        kwargs
        ------
        keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Returns:
            string: Serialized version of object
        """
        json_string = json.dumps(self.to_dict(**kwargs), cls=encoder)
        if uri:
            self._write_to_file(s=json_string, uri=uri, **kwargs)
        return json_string

    @classmethod
    def from_json(
        cls,
        json_string: str = None,
        uri: str = None,
        decoder: callable = json.JSONDecoder,
        **kwargs,
    ):
        """Creates an object from JSON string provided or from URI location containing JSON string

        Parameters
        ----------
        json_string: (string, optional)
            JSON string. Defaults to None.
        uri: (string, optional)
            URI location of file containing JSON string. Defaults to None.
        decoder: (callable, optional)
            Custom decoder. Defaults to simple JSON decoder.
        kwargs
        ------
        keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Raises
        ------
        ValueError
            Raised if neither string nor uri is provided

        Returns
        -------
        cls
            Returns instance of the class
        """
        if json_string:
            return cls.from_dict(json.loads(json_string, cls=decoder))
        if uri:
            json_dict = json.loads(cls._read_from_file(uri, **kwargs), cls=decoder)
            return cls.from_dict(json_dict)
        raise ValueError("Must provide either JSON string or URI location")

    def to_yaml(self, uri: str = None, dumper: callable = dumper, **kwargs) -> str:
        """Returns object serialized as a YAML string

        Parameters
        ----------
        uri: (string, optional)
            URI location to save the YAML string. Defaults to None.
        dumper: (callable, optional)
            Custom YAML Dumper. Defaults to CDumper/SafeDumper.

        kwargs
        ------
        side_effect: Optional[SideEffect]
            side effect to take on the dictionary. The side effect can be either
            convert the dictionary keys to "lower" (SideEffect.CONVERT_KEYS_TO_LOWER.value)
            or "upper"(SideEffect.CONVERT_KEYS_TO_UPPER.value) cases.
        keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Returns:
            Union[str, None]
                Serialized version of object.
                None in case when `uri` provided.
        """
        yaml_string = yaml.dump(self.to_dict(**kwargs), Dumper=dumper)
        if uri:
            self._write_to_file(s=yaml_string, uri=uri, **kwargs)
            return None

        return yaml_string

    @classmethod
    def from_yaml(
        cls,
        yaml_string: str = None,
        uri: str = None,
        loader: callable = loader,
        **kwargs,
    ):
        """Creates an object from YAML string provided or from URI location containing YAML string

        Parameters
        ----------
            yaml_string (string, optional): YAML string. Defaults to None.
            uri (string, optional): URI location of file containing YAML string. Defaults to None.
            loader (callable, optional): Custom YAML loader. Defaults to CLoader/SafeLoader.
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Raises
        ------
        ValueError
            Raised if neither string nor uri is provided

        Returns
        -------
        cls
            Returns instance of the class
        """
        if yaml_string:
            return cls.from_dict(yaml.load(yaml_string, Loader=loader))
        if uri:
            yaml_dict = yaml.load(cls._read_from_file(uri=uri, **kwargs), Loader=loader)
            return cls.from_dict(yaml_dict)
        raise ValueError("Must provide either YAML string or URI location")

    @classmethod
    def from_string(
        cls,
        obj_string: str = None,
        uri: str = None,
        loader: callable = loader,
        **kwargs,
    ) -> "Serializable":
        """Creates an object from string provided or from URI location containing string

        Parameters
        ----------
        obj_string: (str, optional)
            String representing the object
        uri: (str, optional)
            URI location of file containing string. Defaults to None.
        loader: (callable, optional)
            Custom YAML loader. Defaults to CLoader/SafeLoader.
        kwargs: (dict)
            keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Returns
        -------
        Serializable
            A Serializable instance
        """
        return cls.from_yaml(yaml_string=obj_string, uri=uri, loader=loader, **kwargs)

    def __repr__(self):
        """Returns printable version of object.

        Parameters
        ----------
        string
            Serialized version of object as a YAML string
        """
        return self.to_yaml()


class DataClassSerializable(Serializable):
    """Wrapper class that inherit from Serializable class.

    Methods
    -------
    to_dict(self) -> dict
        Serializes the object into a dictionary.
    from_dict(cls, obj_dict) -> cls
        Returns an instance of the class instantiated from the dictionary provided.
    """

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        """validate the dictionary.

        Parameters
        ----------
        obj_dict: (dict)
            Dictionary representation of the object

        Returns
        -------
        bool
            True if the validation passed, else False.
        """
        pass

    def to_dict(self, **kwargs) -> Dict:
        """Serializes instance of class into a dictionary

        kwargs
        ------
        side_effect: Optional[SideEffect]
            side effect to take on the dictionary. The side effect can be either
            convert the dictionary keys to "lower" (SideEffect.CONVERT_KEYS_TO_LOWER.value)
            or "upper"(SideEffect.CONVERT_KEYS_TO_UPPER.value) cases.

        Returns
        -------
        Dict
            A dictionary.
        """
        obj_dict = dataclasses.asdict(self)
        if "side_effect" in kwargs and kwargs["side_effect"]:
            obj_dict = DataClassSerializable._normalize_dict(
                obj_dict=obj_dict, case=kwargs["side_effect"]
            )
        return obj_dict

    @classmethod
    def from_dict(
        cls,
        obj_dict: dict,
        side_effect: Optional[SideEffect] = SideEffect.CONVERT_KEYS_TO_LOWER.value,
    ) -> "DataClassSerializable":
        """Returns an instance of the class instantiated by the dictionary provided.

        Parameters
        ----------
        obj_dict: (dict)
            Dictionary representation of the object
        side_effect: Optional[SideEffect]
            side effect to take on the dictionary. The side effect can be either
            convert the dictionary keys to "lower" (SideEffect.CONVERT_KEYS_TO_LOWER.value)
            or "upper"(SideEffect.CONVERT_KEYS_TO_UPPER.value) cases.

        Returns
        -------
        DataClassSerializable
            A DataClassSerializable instance.
        """
        assert obj_dict, "`obj_dict` must not be None."
        if not isinstance(obj_dict, dict):
            raise TypeError("`obj_dict` must be a dictionary.")
        # check if dict not is None and not empty and type is dict
        cls._validate_dict(obj_dict=obj_dict)
        if side_effect:
            obj_dict = cls._normalize_dict(obj_dict, case=side_effect)

        allowed_fields = set([f.name for f in dataclasses.fields(cls)])
        wrong_fields = set(obj_dict.keys()) - allowed_fields
        if wrong_fields:
            logger.warning(
                f"The class {cls.__name__} doesn't contain attributes: `{list(wrong_fields)}`. "
                "These fields will be ignored."
            )

        obj = cls(**{key: obj_dict[key] for key in allowed_fields})

        for key, value in obj_dict.items():
            if isinstance(value, dict) and hasattr(
                getattr(cls(), key).__class__, "from_dict"
            ):
                attribute = getattr(cls(), key).__class__.from_dict(value)
                setattr(obj, key, attribute)
        return obj

    @staticmethod
    def _normalize_dict(
        obj_dict: Dict, case: str = SideEffect.CONVERT_KEYS_TO_LOWER.value
    ) -> Dict:
        """lower all the keys.

        Parameters
        ----------
        obj_dict: (Dict)
            Dictionary representation of the object.
        case: (optional, str). Defaults to "lower".
            the case to normalized to. can be either "lower" (SideEffect.CONVERT_KEYS_TO_LOWER.value)
            or "upper"(SideEffect.CONVERT_KEYS_TO_UPPER.value).

        Returns
        -------
        Dict
            Dictionary representation of the object.
        """
        normalized_obj_dict = {}
        for key, value in obj_dict.items():
            if isinstance(value, dict):
                value = DataClassSerializable._normalize_dict(
                    value, case=SideEffect.CONVERT_KEYS_TO_UPPER.value
                )
            normalized_obj_dict = DataClassSerializable._normalize_key(
                normalized_obj_dict=normalized_obj_dict, key=key, value=value, case=case
            )
        return normalized_obj_dict

    @staticmethod
    def _normalize_key(
        normalized_obj_dict: Dict, key: str, value: Union[str, Dict], case: str
    ) -> Dict:
        """helper function to normalize the key in the case specified and add it back to the dictionary.

        Paramaters
        ----------
        normalized_obj_dict: (Dict)
            the dictionary to append the key and value to.
        key: (str)
            key to be normalized.
        value: (Union[str, Dict])
            value to be added.
        case: (str)
            the case to normalized to. can be either "lower" (SideEffect.CONVERT_KEYS_TO_LOWER.value)
            or "upper"(SideEffect.CONVERT_KEYS_TO_UPPER.value).

        Raises
        ------
        NotImplementedError: if case provided is not either "lower" or "upper".

        Returns
        -------
        Dict
            normalized dictionary with the key and value added in the case specified.
        """
        if case.lower() == SideEffect.CONVERT_KEYS_TO_LOWER.value:
            normalized_obj_dict[key.lower()] = value
        elif case.lower() == SideEffect.CONVERT_KEYS_TO_UPPER.value:
            normalized_obj_dict[key.upper()] = value
        else:
            raise NotImplementedError("`case` is not supported.")
        return normalized_obj_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that helps to manage feature types.
Provides functionalities to register, unregister, list feature types.

Classes
--------
    FeatureTypeManager
        Feature Types Manager class that manages feature types.

Examples
--------
    >>> from ads.feature_engineering.feature_type.base import FeatureType
    >>> class NewType(FeatureType):
    ...    description="My personal type."
    ...    pass
    >>> FeatureTypeManager.feature_type_register(NewType)
    >>> FeatureTypeManager.feature_type_registered()
                Name        Feature Type                                  Description
    ---------------------------------------------------------------------------------
    0	  Continuous          continuous          Type representing continuous values.
    1	    DateTime           date_time           Type representing date and/or time.
    2	    Category            category  Type representing discrete unordered values.
    3	     Ordinal             ordinal             Type representing ordered values.
    4        NewType            new_type                             My personal type.

    >>> FeatureTypeManager.warning_registered()
        Feature Type             Warning                    Handler
    ----------------------------------------------------------------------
    0     continuous               zeros              zeros_handler
    1     continuous    high_cardinality   high_cardinality_handler

    >>> FeatureTypeManager.validator_registered()
        Feature Type            Validator                 Condition                     Handler
    -------------------------------------------------------------------------------------------
    0   phone_number      is_phone_number                        ()             default_handler
    1   phone_number      is_phone_number    {'country_code': '+7'}    specific_country_handler
    2    credit_card       is_credit_card                        ()             default_handler

    >>> FeatureTypeManager.feature_type_unregister(NewType)
    >>> FeatureTypeManager.feature_type_reset()
    >>> FeatureTypeManager.feature_type_object('continuous')
    Continuous
"""
from typing import Union
import logging
import pandas as pd
import pandas.api.types as pdtypes
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.feature_type.object import Object
from ads.feature_engineering.feature_type.integer import Integer
from ads.feature_engineering.feature_type.category import Category
from ads.feature_engineering.feature_type.ordinal import Ordinal
from ads.feature_engineering.feature_type.boolean import Boolean
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.feature_type.lat_long import LatLong
from ads.feature_engineering.feature_type.creditcard import CreditCard
from ads.feature_engineering.feature_type.zip_code import ZipCode
from ads.feature_engineering.feature_type.phone_number import PhoneNumber
from ads.feature_engineering.feature_type.datetime import DateTime
from ads.feature_engineering.feature_type.continuous import Continuous
from ads.feature_engineering.feature_type.address import Address
from ads.feature_engineering.feature_type.constant import Constant
from ads.feature_engineering.feature_type.document import Document
from ads.feature_engineering.feature_type.gis import GIS
from ads.feature_engineering.feature_type.ip_address import IpAddress
from ads.feature_engineering.feature_type.ip_address_v4 import IpAddressV4
from ads.feature_engineering.feature_type.ip_address_v6 import IpAddressV6
from ads.feature_engineering.feature_type.text import Text
from ads.feature_engineering.feature_type.unknown import Unknown
from ads.feature_engineering.feature_type.discrete import Discrete
from ads.feature_engineering import exceptions
from ads.feature_engineering.feature_type.adsstring.string import ADSString

logger = logging.getLogger(__name__)


def _feature_type_by_dtype(dtype) -> FeatureType:
    """Determines feature type by DataFrame dtype.

    Parameters
    ----------
    dtype: pd.DataFrame.dtypes
        The Pandas series data type.

    Returns
    -------
    FeatureType
        The subclass of FeatureType.
    """
    if pdtypes.is_bool_dtype(dtype):
        return Boolean
    if pdtypes.is_datetime64_any_dtype(dtype):
        return DateTime
    if pdtypes.is_categorical_dtype(dtype):
        return Category
    if pdtypes.is_string_dtype(dtype):
        return String
    if pdtypes.is_float_dtype(dtype):
        return Continuous
    if pdtypes.is_integer_dtype(dtype):
        return Integer
    return Object


class FeatureTypeManager:
    """Feature Types Manager class that manages feature types.

    Provides functionalities to register, unregister, list feature types.

    Methods
    -------
    feature_type_object(cls, feature_type: Union[FeatureType, str]) -> FeatureType
        Gets a feature type by class object or name.
    feature_type_register(cls, feature_type_cls: FeatureType) -> None
        Registers a feature type.
    feature_type_unregister(cls, feature_type_cls: Union[FeatureType, str]) -> None
        Unregisters a feature type.
    feature_type_reset(cls) -> None
        Resets feature types to be default.
    feature_type_registered(cls) -> pd.DataFrame
        Lists all registered feature types as a DataFrame.
    warning_registered(cls) -> pd.DataFrame
        Lists registered warnings for all registered feature types.
    validator_registered(cls) -> pd.DataFrame
        Lists registered validators for all registered feature types.

    Examples
    --------
    >>> from ads.feature_engineering.feature_type.base import FeatureType
    >>> class NewType(FeatureType):
    ...    pass
    >>> FeatureTypeManager.register_feature_type(NewType)
    >>> FeatureTypeManager.feature_type_registered()
                Name      Feature Type                                  Description
    -------------------------------------------------------------------------------
    0	  Continuous        continuous          Type representing continuous values.
    1	    DateTime         date_time           Type representing date and/or time.
    2	    Category          category  Type representing discrete unordered values.
    3	     Ordinal           ordinal             Type representing ordered values.

    >>> FeatureTypeManager.warning_registered()
        Feature Type             Warning                    Handler
    ----------------------------------------------------------------------
    0     continuous               zeros              zeros_handler
    1     continuous    high_cardinality   high_cardinality_handler

    >>> FeatureTypeManager.validator_registered()
        Feature Type            Validator                 Condition                     Handler
    -------------------------------------------------------------------------------------------
    0   phone_number      is_phone_number                        ()             default_handler
    1   phone_number      is_phone_number    {'country_code': '+7'}    specific_country_handler
    2    credit_card       is_credit_card                        ()             default_handler

    >>> FeatureTypeManager.feature_type_unregister(NewType)
    >>> FeatureTypeManager.feature_type_reset()
    >>> FeatureTypeManager.feature_type_object('continuous')
    Continuous
    """

    _default_registered_type = [
        Continuous,
        DateTime,
        Category,
        Ordinal,
        Boolean,
        String,
        LatLong,
        PhoneNumber,
        ZipCode,
        CreditCard,
        Object,
        Integer,
        Address,
        Constant,
        Document,
        GIS,
        IpAddressV4,
        IpAddressV6,
        IpAddress,
        Text,
        Unknown,
        Discrete,
        ADSString,
    ]
    _name_to_type_map = {tp.name: tp for tp in _default_registered_type}

    @classmethod
    def feature_type_register(cls, feature_type_cls: FeatureType) -> None:
        """Registers new feature type.

        Parameters
        ----------
        feature_type : FeatureType
            Subclass of FeatureType to be registered.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            Type is not a subclass of FeatureType.
        TypeError
            Type has already been registered.
        NameError
            Name has already been used.
        """
        if not issubclass(feature_type_cls, FeatureType):
            raise exceptions.InvalidFeatureType(feature_type_cls.__name__)
        if feature_type_cls in cls._name_to_type_map.values():
            raise exceptions.TypeAlreadyRegistered(feature_type_cls.__name__)
        if feature_type_cls.name in cls._name_to_type_map:
            raise exceptions.NameAlreadyRegistered(feature_type_cls.name)

        cls._name_to_type_map[feature_type_cls.name] = feature_type_cls

    @classmethod
    def feature_type_unregister(cls, feature_type: Union[FeatureType, str]) -> None:
        """Unregisters a feature type.

        Parameters
        ----------
        feature_type: (FeatureType | str)
            The FeatureType subclass or a str indicating feature type.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            In attempt to unregister a default feature type.
        """
        feature_type_cls = cls.feature_type_object(feature_type)
        if feature_type_cls in cls._default_registered_type:
            raise TypeError(
                f"Default type {feature_type_cls.__name__} cannot be removed."
            )

        del cls._name_to_type_map[feature_type_cls.name]

    @classmethod
    def feature_type_reset(cls) -> None:
        """Resets feature types to be default.

        Returns
        -------
        None
            Nothing.
        """
        cls._name_to_type_map = {tp.name: tp for tp in cls._default_registered_type}

    @classmethod
    def feature_type_registered(cls) -> pd.DataFrame:
        """Lists all registered feature types as a DataFrame.

        Returns
        -------
        pd.DataFrame:
            The list of feature types in a DataFrame format.
        """
        return (
            pd.DataFrame(
                [
                    (ft.__name__, name, ft.description)
                    for name, ft in cls._name_to_type_map.items()
                ],
                columns=["Class", "Name", "Description"],
            )
            .sort_values(by=["Class", "Name"])
            .reset_index(drop=True)
        )

    @classmethod
    def feature_type_object(cls, feature_type: Union[FeatureType, str]) -> FeatureType:
        """Gets a feature type by class object or name.

        Parameters
        ----------
        feature_type: Union[FeatureType, str]
            The FeatureType subclass or a str indicating feature type.

        Returns
        -------
        FeatureType
            Found feature type.

        Raises
        ------
        TypeNotFound
            If provided feature type not registered.
        TypeError
            If provided feature type not a subclass of FeatureType.
        """

        if isinstance(feature_type, str):
            if feature_type not in cls._name_to_type_map:
                raise exceptions.TypeNotFound(feature_type)
            return cls._name_to_type_map[feature_type]

        if not isinstance(feature_type, FeatureType) and not issubclass(
            feature_type, FeatureType
        ):
            raise TypeError(
                f"{feature_type} must be an instance or subclass of FeatureType."
            )

        if isinstance(feature_type, FeatureType):
            feature_type_cls = feature_type.__class__
        else:
            feature_type_cls = feature_type

        if feature_type_cls not in cls._name_to_type_map.values():
            raise exceptions.TypeNotFound(feature_type_cls.__name__)

        return feature_type_cls

    @classmethod
    def is_type_registered(cls, feature_type: Union[FeatureType, str]) -> bool:
        """Checks if provided feature type registered in the system.

        Parameters
        ----------
        feature_type: Union[FeatureType, str]
            The FeatureType subclass or a str indicating feature type.

        Returns
        -------
        bool
            True if provided feature type registered, False otherwise.
        """
        result = False
        try:
            cls.feature_type_object(feature_type)
        except Exception as e:  # pylint: disable=broad-except
            logger.info("Error %s occured. Arguments %s", str(e), e.args)
        else:
            result = True
        return result

    @classmethod
    def warning_registered(cls) -> pd.DataFrame:
        """Lists registered warnings for all registered feature types.

        Returns
        -------
        pd.DataFrame:
            The list of registered warnings for registered feature types
            in a DataFrame format.

        Examples
        --------
        >>> FeatureTypeManager.warning_registered()
            Feature Type             Warning                    Handler
        ----------------------------------------------------------------------
        0     continuous               zeros              zeros_handler
        1     continuous    high_cardinality   high_cardinality_handler
        """
        result_df = pd.DataFrame((), columns=["Feature Type", "Warning", "Handler"])
        for feature_type in cls._name_to_type_map.values():
            feature_type_df = feature_type.warning.registered()
            feature_type_df.insert(0, "Feature Type", feature_type.name)
            feature_type_df = feature_type_df.rename(columns={"Name": "Warning"})
            result_df = result_df.append(feature_type_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df

    @classmethod
    def validator_registered(cls) -> pd.DataFrame:
        """Lists registered validators for registered feature types.

        Returns
        -------
        pd.DataFrame:
            The list of registered validators for registered feature types
            in a DataFrame format.

        Examples
        --------
        >>> FeatureTypeManager.validator_registered()
            Feature Type            Validator                 Condition                     Handler
        -------------------------------------------------------------------------------------------
        0   phone_number      is_phone_number                        ()             default_handler
        1   phone_number      is_phone_number    {'country_code': '+7'}    specific_country_handler
        2    credit_card       is_credit_card                        ()             default_handler
        """
        result_df = pd.DataFrame(
            [], columns=["Feature Type", "Validator", "Condition", "Handler"]
        )
        for feature_type in cls._name_to_type_map.values():
            feature_type_df = feature_type.validator.registered()
            feature_type_df.insert(0, "Feature Type", feature_type.name)
            feature_type_df = feature_type_df.rename(columns={"Name": "Validator"})
            result_df = result_df.append(feature_type_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import cloudpickle
import numpy as np
import pandas as pd
import base64
import os
import json
from io import BytesIO
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import logger
from ads.model.serde.common import Serializer, Deserializer
from functools import lru_cache
from pickle import UnpicklingError
from typing import Dict, List, Union

SUPPORTED_MODEL_INPUT_SERIALIZERS = ["json", "cloudpickle", "spark"]


class ModelInputSerializerType:
    JSON = "json"
    CLOUDPICKLE = "cloudpickle"


class SparkModelInputSerializerType:
    SPARK = "spark"


class ModelInputSerializer(Serializer):
    """Abstract base class for creation of new data serializers."""

    def __init__(self):
        super().__init__()

    def serialize(self, data):
        return data


class JsonModelInputSerializer(ModelInputSerializer):
    """
    ADS data serializer. Serialize data of various formats to into a
    dictionary containing serialized input data and original data type
    information.


    Examples
    --------
    >>> from ads.model.serde.model_input import JsonModelInputSerializer

    >>> # numpy array will be converted to base64 encoded string,
    >>> # while `data_type` will record its original type: `numpy.ndarray`
    >>> import numpy as np
    >>> input_data = np.array([1, 2, 3])
    >>> serialized_data = JsonModelInputSerializer().serialize(data=input_data)
    >>> serialized_data
    {
    'data': 'k05VTVBZAQB2AHsnZGVzY3InOiAnPGk4JywgJ2ZvcnRyYW5fb3JkZXInOiBGYWxzZS
    wgJ3NoYXBlJzogKDMsKSwgfSAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgI
    CAgICAgICAgICAgICAgICAgICAgIAoBAAAAAAAAAAIAAAAAAAAAAwAAAAAAAAA=',
    'data_type': "<class 'numpy.ndarray'>"
    }

    >>> # `pd.core.frame.DataFrame` will be converted to json by `.to_json()`
    >>> # while `data_type` will record its original type: `pandas.core.frame.DataFrame`
    >>> import pandas as pd
    >>> df = pd.DataFrame(data={'col1': [1, 2], 'col2': [3, 4]})
    >>> serialized_data = JsonModelInputSerializer().serialize(data=df)
    >>> serialized_data
    {
    'data': '{"col1":{"0":1,"1":2},"col2":{"0":3,"1":4}}',
    'data_type': "<class 'pandas.core.frame.DataFrame'>"
    }

    >>> # `pandas.core.series.Series` will be converted to list by `.tolist()`
    >>> # while `data_type` will record its original type: `pandas.core.series.Series`
    >>> ser = pd.Series(data={'a': 1, 'b': 2, 'c': 3}, index=['a', 'b', 'c'])
    >>> serialized_data = JsonModelInputSerializer().serialize(data=ser)
    >>> serialized_data
    {
    'data': [1, 2, 3],
    'data_type': "<class 'pandas.core.series.Series'>"
    }

    >>> # `torch.Tensor` will be converted to base64 encoded string,
    >>> # while `data_type` will record its original type: `torch.Tensor`
    >>> import torch
    >>> tt = torch.tensor([[1, 2, 3], [4, 5, 6]])
    >>> serialized_data = JsonModelInputSerializer().serialize(data=tt)
    >>> serialized_data
    {
    'data': 'UEsDBAAACAgAAAAAAAAAAAAAAAAAAAAAAAAQABIAYXJjaGl2ZS9kYXRhLnBrbEZCDgBaWlpaW
    lpaWlpaWlpaWoACY3RvcmNoLl91dGlscwpfcmVidWlsZF90ZW5zb3JfdjIKcQAoKFgHAAAAc3RvcmFnZXEB
    Y3RvcmNoCkxvbmdTdG9yYWdlCnECWAEAAAAwcQNYAwAAAGNwdXEESwZ0cQVRSwBLAksDhnEGSwNLAYZxB4l
    jY29sbGVjdGlvbnMKT3JkZXJlZERpY3QKcQgpUnEJdHEKUnELLlBLBwim2iAhmQAAAJkAAABQSwMEAAAICA
    AAAAAAAAAAAAAAAAAAAAAAAA4AKwBhcmNoaXZlL2RhdGEvMEZCJwBaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWl
    paWlpaWlpaWlpaWlpaWloBAAAAAAAAAAIAAAAAAAAAAwAAAAAAAAAEAAAAAAAAAAUAAAAAAAAABgAAAAAAA
    ABQSwcI9z/uVjAAAAAwAAAAUEsDBAAACAgAAAAAAAAAAAAAAAAAAAAAAAAPABMAYXJjaGl2ZS92ZXJzaW9u
    RkIPAFpaWlpaWlpaWlpaWlpaWjMKUEsHCNGeZ1UCAAAAAgAAAFBLAQIAAAAACAgAAAAAAACm2iAhmQAAAJk
    AAAAQAAAAAAAAAAAAAAAAAAAAAABhcmNoaXZlL2RhdGEucGtsUEsBAgAAAAAICAAAAAAAAPc/7lYwAAAAMA
    AAAA4AAAAAAAAAAAAAAAAA6QAAAGFyY2hpdmUvZGF0YS8wUEsBAgAAAAAICAAAAAAAANGeZ1UCAAAAAgAAA
    A8AAAAAAAAAAAAAAAAAgAEAAGFyY2hpdmUvdmVyc2lvblBLBgYsAAAAAAAAAB4DLQAAAAAAAAAAAAMAAAAA
    AAAAAwAAAAAAAAC3AAAAAAAAANIBAAAAAAAAUEsGBwAAAACJAgAAAAAAAAEAAABQSwUGAAAAAAMAAwC3AAA
    A0gEAAAAA',
    'data_type': "<class 'torch.Tensor'>"
    }

    >>> # `tensorflow.Tensor` will be converted to base64 encoded string,
    >>> # while `data_type` will record its original type: `tensorflow.python.framework.ops.EagerTensor`.
    >>> import torch
    >>> c = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    >>> serialized_data = JsonModelInputSerializer().serialize(data=c)
    >>> serialized_data
    {
    'data': 'k05VTVBZAQB2AHsnZGVzY3InOiAnPGY0JywgJ2ZvcnRyYW5fb3JkZXInOiBGYWxzZSwgJ3NoYXB
    lJzogKDIsIDIpLCB9ICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA
    gICAgICAgIAoAAIA/AAAAQAAAQEAAAIBA',
    'data_type': "<class 'tensorflow.python.framework.ops.EagerTensor'>"
    }

    >>> # dict, str, list, tuple will be saved as orignal type
    >>> # and `data_type` will record its type.
    >>> mystring = "this is a string."
    >>> serialized_data = JsonModelInputSerializer().serialize(data=mystring)
    >>> serialized_data
    {
    'data': 'this is a string.',
    'data_type': "<class 'str'>"
    }
    """

    def __init__(self):
        super().__init__()

    @runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
    def _convert_torch_tensor(self, data):
        buffer = BytesIO()
        torch.save(data, buffer)
        data = base64.b64encode(buffer.getvalue()).decode("utf-8")
        return data

    @runtime_dependency(
        module="tensorflow",
        short_name="tf",
        install_from=OptionalDependency.TENSORFLOW,
    )
    def _convert_tf_tensor(self, data):
        data = data.numpy()
        return data

    def serialize(
        self,
        data: Union[
            Dict,
            str,
            List,
            np.ndarray,
            pd.core.series.Series,
            pd.core.frame.DataFrame,
        ],
    ):
        """Serialize data into a dictionary containing serialized input
        data and original data type information.

        Parameters
        ----------
        data: Union[Dict, str, list, numpy.ndarray, pd.core.series.Series,
        pd.core.frame.DataFrame, bytes]
            Data expected by the model deployment predict API.

        Returns
        -------
        Dict
            A dictionary containing serialized input data and original data type
            information.

        Raises
        ------
        TypeError
            if provided data type is not supported.
        """
        data_type = str(type(data))
        if "torch.Tensor" in data_type:
            data = self._convert_torch_tensor(data)
        if "tensorflow.python.framework.ops.EagerTensor" in data_type:
            data = self._convert_tf_tensor(data)

        if isinstance(data, np.ndarray):
            np_bytes = BytesIO()
            np.save(np_bytes, data, allow_pickle=True)
            data = base64.b64encode(np_bytes.getvalue()).decode("utf-8")
        elif isinstance(data, pd.core.series.Series):
            data = data.tolist()
        elif isinstance(data, pd.core.frame.DataFrame):
            data = data.to_json()
        elif (
            isinstance(data, dict)
            or isinstance(data, str)
            or isinstance(data, list)
            or isinstance(data, tuple)
            or isinstance(data, bytes)
        ):
            pass
        else:
            raise TypeError(
                "The supported data types are Dict, str, list, bytes, tuple, "
                "numpy.ndarray, pd.core.series.Series, tf.Tensor, torch.Tensor, "
                "pd.core.frame.DataFrame. Please convert to the supported data "
                "types first. "
            )

        data_dict = {
            "data": data,
            "data_type": data_type,
        }
        return data_dict


class CloudpickleModelInputSerializer(ModelInputSerializer):
    """Serialize data of various formats to bytes."""

    def __init__(self):
        super().__init__()

    def serialize(self, data):
        """Serialize data into bytes.

        Parameters
        ----------
        data (object): Data to be serialized.

        Returns
        -------
        object: Serialized data used for a request.
        """
        serialized_data = cloudpickle.dumps(data)
        return serialized_data


class SparkModelInputSerializer(JsonModelInputSerializer):
    """[An internal class]
    Defines the contract for input data to spark pipeline models.

    """

    def __init__(self):
        super().__init__()

    @runtime_dependency(
        module="pyspark",
        short_name="sql",
        object="sql",
        install_from=OptionalDependency.SPARK,
    )
    def serialize(
        self,
        data: Union[
            Dict,
            str,
            List,
            np.ndarray,
            pd.core.series.Series,
            pd.core.frame.DataFrame,
        ],
    ):
        """
        Parameters
        ----------
        data: Union[Dict, str, list, numpy.ndarray, pd.core.series.Series,
        pd.core.frame.DataFrame]
            Data expected by the model deployment predict API.

        """
        data, _, _ = self._serialize_via_spark(data)
        if isinstance(data, sql.DataFrame):
            data = data.toJSON().collect()

        try:
            data = super().serialize(data=data)
        except:
            raise TypeError(
                f"Data type: {type(data)} unsupported. Please use `pyspark.sql.DataFrame`, `pyspark.pandas.DataFrame`, `pandas.DataFrame`."
            )
        return data

    @runtime_dependency(
        module="pyspark",
        short_name="sql",
        object="sql",
        install_from=OptionalDependency.SPARK,
    )
    def _serialize_via_spark(self, data):
        """
        If data is either a spark SQLDataFrames and spark.pandas dataframe/series
            Return pandas version and data type of original
        Else
            Return data and None
        """
        try:  # runtime_dependency could not import this for unknown reason
            import pyspark.pandas as ps

            ps_available = True
        except:
            ps_available = False

        def _get_or_create_spark_session():
            return sql.SparkSession.builder.appName(
                "Convert pandas to spark"
            ).getOrCreate()

        if isinstance(data, sql.DataFrame):
            data_type = type(data)
        elif ps_available and (
            isinstance(data, ps.DataFrame) or isinstance(data, ps.Series)
        ):
            data_type = type(data)
            data = data.to_spark()
        elif isinstance(data, sql.types.Row):
            spark_session = _get_or_create_spark_session()
            data = spark_session.createDataFrame(data)
            data_type = type(data)
        elif isinstance(data, pd.core.frame.DataFrame):
            data_type = type(data)
            spark_session = _get_or_create_spark_session()
            data = spark_session.createDataFrame(data)
        elif isinstance(data, list):
            if not len(data):
                raise TypeError(
                    f"Data cannot be empty. Provided data parameter is: {data}"
                )
            if isinstance(data[0], sql.types.Row):
                spark_session = _get_or_create_spark_session()
                data = spark_session.createDataFrame(data)
                data_type = type(data)
            else:
                logger.warn(
                    f"ADS does not serialize data type: {type(data)} for Spark Models. User should proceed at their own risk. ADS supported data types are: `pyspark.sql.DataFrame`, `pandas.DataFrame`, and `pyspark.pandas.DataFrame`."
                )
                return data, type(data), None
        else:
            logger.warn(
                f"ADS does not serialize data type: {type(data)} for Spark Models. User should proceed at their own risk. ADS supported data types are: `pyspark.sql.DataFrame`, `pandas.DataFrame`, and `pyspark.pandas.DataFrame`."
            )
            return data, type(data), None
        return data, data_type, data.schema


class ModelInputDeserializer(Deserializer):
    """Abstract base class for creation of new data deserializers."""

    def __init__(self, name="customized"):
        super().__init__()
        self.name = name

    def deserialize(self, data):
        return data


class JsonModelInputDeserializer(ModelInputDeserializer):
    """ADS data deserializer. Deserialize data to into its original type."""

    def __init__(self, name="json"):
        super().__init__(name=name)

    @lru_cache(maxsize=1)
    def _fetch_data_type_from_schema(
        input_schema_path=os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "input_schema.json"
        )
    ):
        """
        Returns data type information fetch from input_schema.json.

        Parameters
        ----------
        input_schema_path: path of input schema.

        Returns
        -------
        data_type: data type fetch from input_schema.json.

        """
        data_type = {}
        if os.path.exists(input_schema_path):
            schema = json.load(open(input_schema_path))
            for col in schema["schema"]:
                data_type[col["name"]] = col["dtype"]
        else:
            print(
                "input_schema has to be passed in in order to recover the same data type. pass `X_sample` in `ads.model.framework.pytorch_model.PyTorchModel.prepare` function to generate the input_schema. Otherwise, the data type might be changed after serialization/deserialization."
            )
        return data_type

    def deserialize(self, data: dict):
        """Deserialize data into its original type.

        Parameters
        ----------
        data: (dict)

        Returns
        -------
        objects: Deserialized data used for a prediction.

        Raises
        ------
        TypeError
            if provided data type is not supported.
        """
        if isinstance(data, bytes):
            logger.warning(
                "bytes are passed directly to the model. If the model expects a specific data format, you need to write the conversion logic in `deserialize()` yourself."
            )
            return data

        data_type = data.get("data_type", "") if isinstance(data, dict) else ""
        json_data = data.get("data", data) if isinstance(data, dict) else data
        if "torch.Tensor" in data_type:
            return self._load_torch_tensor(json_data)
        if "tensorflow.python.framework.ops.EagerTensor" in data_type:
            return self._load_tf_tensor(json_data)
        if "numpy.ndarray" in data_type:
            load_bytes = BytesIO(base64.b64decode(json_data.encode("utf-8")))
            return np.load(load_bytes, allow_pickle=True)
        if "pandas.core.series.Series" in data_type:
            return pd.Series(json_data)
        if "pandas.core.frame.DataFrame" in data_type or isinstance(json_data, str):
            return pd.read_json(json_data, dtype=self._fetch_data_type_from_schema())
        if isinstance(json_data, dict):
            return pd.DataFrame.from_dict(json_data)

        return json_data

    @runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
    def _load_torch_tensor(self, data):
        load_bytes = BytesIO(base64.b64decode(data.encode("utf-8")))
        return torch.load(load_bytes)

    @runtime_dependency(
        module="tensorflow",
        short_name="tf",
        install_from=OptionalDependency.TENSORFLOW,
    )
    def _load_tf_tensor(self, data):
        load_bytes = BytesIO(base64.b64decode(data.encode("utf-8")))
        return tf.convert_to_tensor(np.load(load_bytes, allow_pickle=True))


class SparkModelInputDeserializer(ModelInputDeserializer):
    def __init__(self, name="spark"):
        super().__init__(name=name)

    def deserialize(data):
        """
        Not implement. See spark template.
        """
        pass


class CloudpickleModelInputDeserializer(ModelInputDeserializer):
    """Use cloudpickle to deserialize data into its original type."""

    def __init__(self, name="cloudpickle"):
        super().__init__(name=name)

    def deserialize(self, data):
        """Deserialize data into its original type.

        Parameters
        ----------
        data (object): Data to be deserialized.

        Returns
        -------
        object: deserialized data used for a prediction.
        """
        deserialized_data = data
        try:
            deserialized_data = cloudpickle.loads(data)
        except TypeError:
            pass
        except UnpicklingError:
            logger.warning(
                "bytes are passed directly to the model. If the model expects a specific data format, you need to write the conversion logic in `deserialize()` yourself."
            )
        return deserialized_data


class JsonModelInputSERDE(JsonModelInputSerializer, JsonModelInputDeserializer):
    name = "json"


class CloudpickleModelInputSERDE(
    CloudpickleModelInputSerializer, CloudpickleModelInputDeserializer
):
    name = "cloudpickle"


class SparkModelInputSERDE(SparkModelInputSerializer, SparkModelInputDeserializer):
    name = "spark"


class ModelInputSerializerFactory:
    """Data Serializer Factory.

    Examples
    --------
    >>> serializer, deserializer = ModelInputSerializerFactory.get("cloudpickle")
    """

    _factory = {}
    _factory["json"] = JsonModelInputSERDE
    _factory["cloudpickle"] = CloudpickleModelInputSERDE
    _factory["spark"] = SparkModelInputSERDE

    @classmethod
    def get(cls, se: str = "json"):
        """Gets data serializer and corresponding deserializer.

        Parameters
        ----------
        se (str):
            The name of the required serializer.

        Raises
        ------
        ValueError:
            Raises when input is unsupported format.

        Returns
        -------
        serde (ads.model.serde.common.SERDE):
            Intance of `ads.model.serde.common.SERDE".
        """
        serde_cls = cls._factory.get(se, None)
        if serde_cls:
            return serde_cls()
        else:
            raise ValueError(
                f"This {se} format is not supported."
                f"Currently support the following format: {SUPPORTED_MODEL_INPUT_SERIALIZERS}."
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import itertools
from collections import Counter
from typing import List, Sequence, Tuple
import os
import functools
import pandas as pd

from ads.feature_engineering.adsstring.parsers.base import Parser
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


# This module is only used when user installed spacy. spacy has conflicts with oci-cli.
class SpacyParser(Parser):  # pragma: no cover
    @runtime_dependency(module="spacy", install_from=OptionalDependency.TEXT)
    def __init__(self):
        try:
            # for ADS conda packs, spacy files are located under $CONDA_PREFIX/spacy
            self.nlp = spacy.load(f"{os.environ['CONDA_PREFIX']}/spacy")
        except OSError:
            try:
                # if not inside a conda pack, trying loading first
                # if failed download to default path
                self.nlp = spacy.load("en_core_web_sm")
            except:
                from spacy.cli import download

                download("en_core_web_sm")
                self.nlp = spacy.load("en_core_web_sm")

    @property
    @functools.lru_cache()
    def _parsed_nlp_obj(self):
        return self.nlp(self.string)

    @property
    @functools.lru_cache()
    def pos(self) -> pd.DataFrame:
        pos = (
            (token.text, token.pos_)
            for token in self._parsed_nlp_obj
            if not token.is_space
        )
        return pd.DataFrame(data=pos, columns=["Word", "Label"])

    def _pos_tokens(self, pos: str) -> List[str]:
        return _f7(
            [
                token.text
                for token in self._parsed_nlp_obj
                if (
                    not token.is_stop
                    and not token.is_punct
                    and not token.is_space
                    and token.pos_ == pos
                )
            ]
        )

    @property
    @functools.lru_cache()
    def noun(self) -> List[str]:
        return self._pos_tokens("NOUN")

    @property
    @functools.lru_cache()
    def adjective(self) -> List[str]:
        return self._pos_tokens("ADJ")

    @property
    @functools.lru_cache()
    def adverb(self) -> List[str]:
        return self._pos_tokens("ADV")

    @property
    @functools.lru_cache()
    def verb(self) -> List[str]:
        return self._pos_tokens("VERB")

    @property
    @functools.lru_cache()
    def noun_phrase(self) -> List[str]:
        np = []
        for sentence in self.sentence:
            for chunk in self.nlp(sentence).noun_chunks:
                if chunk.text.strip():
                    np.append(chunk.text.strip())

        return np

    @property
    @functools.lru_cache()
    def entity_extract(self) -> pd.DataFrame:
        return pd.DataFrame(data=self._entity_extract(), columns=["Entity", "Label"])

    @functools.lru_cache()
    def _entity_extract(self) -> List[Tuple[str, str]]:
        entities = []
        for sentence in self.sentence:
            entities.extend([(ent.text, ent.label_) for ent in self.nlp(sentence).ents])

        return _f7(entities)

    def _entity_tokens(self, entity_label: str) -> List[str]:
        return [
            token for (token, label) in self._entity_extract() if label == entity_label
        ]

    @property
    @functools.lru_cache()
    def entity_people(self) -> List[str]:
        # People, including fictional.
        return self._entity_tokens("PERSON")

    @property
    @functools.lru_cache()
    def entity_location(self) -> List[str]:
        # Location entity, i.e. countries, cities, states, mountain ranges, bodies of water, buildings, airports, highways, bridges
        return (
            self._entity_tokens("GPE")
            + self._entity_tokens("LOC")
            + self._entity_tokens("FAC")
        )

    @property
    @functools.lru_cache()
    def entity_organization(self) -> List[str]:
        # Companies, agencies, institutions.
        return self._entity_tokens("ORG")

    @property
    @functools.lru_cache()
    def entity_artwork(self) -> List[str]:
        # title of books, songs, etc
        return self._entity_tokens("WORK_OF_ART")

    @property
    @functools.lru_cache()
    def entity_product(self) -> List[str]:
        # product names, etc
        return self._entity_tokens("PRODUCT")

    @property
    @functools.lru_cache()
    def sentence(self) -> List[str]:
        return [sent.text.strip() for sent in self._parsed_nlp_obj.sents]

    @property
    @functools.lru_cache()
    def token(self) -> List[str]:
        return [token.text for token in self._parsed_nlp_obj]

    def _words_by_sentence(self) -> List[str]:
        x = []
        for sentence in self.sentence:
            x.append([token.text for token in self.nlp(sentence) if token.is_alpha])

        return x

    @property
    @functools.lru_cache()
    def word(self) -> List[str]:
        return list(itertools.chain(*self._words_by_sentence()))

    def _ngrams(self, tokens, n):
        return [tokens[i : i + n] for i in range(len(tokens) - n + 1)]

    @property
    @functools.lru_cache()
    def bigram(self) -> pd.DataFrame:
        bigram = list(
            itertools.chain(
                *[
                    self._ngrams(sentence_words, 2)
                    for sentence_words in self._words_by_sentence()
                ]
            )
        )
        return pd.DataFrame(data=bigram, columns=["Word 1", "Word 2"])

    @property
    @functools.lru_cache()
    def trigram(self) -> pd.DataFrame:
        trigram = list(
            itertools.chain(
                *[
                    self._ngrams(sentence_words, 3)
                    for sentence_words in self._words_by_sentence()
                ]
            )
        )
        return pd.DataFrame(data=trigram, columns=["Word 1", "Word 2", "Word 3"])

    @property
    @functools.lru_cache()
    def lemma(self) -> List[str]:
        return [token.lemma_ for token in self._parsed_nlp_obj]

    @property
    @functools.lru_cache()
    def word_count(self) -> pd.DataFrame:
        word_count = Counter([word.lower() for word in self.word]).most_common()
        df = pd.DataFrame(data=word_count, columns=["Word", "Count"])
        return df.sort_values(
            ["Count", "Word"], ascending=[False, True], ignore_index=True
        )


def _f7(seq: Sequence) -> List:
    """order preserving de-duplicate sequence"""
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import os
import re
import warnings
import oci
import datetime
import pandas as pd
from fsspec.utils import infer_storage_options
import inspect
import fsspec

from ads.common import utils
from ads.common.utils import is_same_class
from ads.dataset import logger
from ads.dataset.classification_dataset import (
    BinaryClassificationDataset,
    MultiClassClassificationDataset,
    BinaryTextClassificationDataset,
    MultiClassTextClassificationDataset,
)
from ads.dataset.dataset import ADSDataset
from ads.dataset.forecasting_dataset import ForecastingDataset
from ads.dataset.helper import (
    get_feature_type,
    is_text_data,
    generate_sample,
    DatasetDefaults,
    ElaboratedPath,
    DatasetLoadException,
)
from ads.dataset.regression_dataset import RegressionDataset
from ads.type_discovery.type_discovery_driver import TypeDiscoveryDriver
from ads.type_discovery.typed_feature import (
    ContinuousTypedFeature,
    DateTimeTypedFeature,
    CategoricalTypedFeature,
    OrdinalTypedFeature,
    GISTypedFeature,
    DocumentTypedFeature,
)
from ads.type_discovery.typed_feature import TypedFeature
from typing import Callable, Tuple
from ocifs import OCIFileSystem
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.decorator.deprecate import deprecated

default_snapshots_dir = None
default_storage_options = None
mindate = datetime.date(datetime.MINYEAR, 1, 1)


class DatasetFactory:
    @staticmethod
    @deprecated(
        "2.6.6",
        details="Deprecated in favor of using Pandas. Pandas supports reading from object storage directly. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/loading_data/connect.html",
    )
    def open(
        source,
        target=None,
        format="infer",
        reader_fn: Callable = None,
        name: str = None,
        description="",
        npartitions: int = None,
        type_discovery=True,
        html_table_index=None,
        column_names="infer",
        sample_max_rows=10000,
        positive_class=None,
        transformer_pipeline=None,
        types={},
        **kwargs,
    ):
        """
        Returns an object of ADSDataset or ADSDatasetWithTarget  read from the given path

        .. deprecated:: 2.6.6
            "Deprecated in favor of using Pandas. Pandas supports reading from object storage directly.
            Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/loading_data/connect.html",

        Parameters
        ----------
        source: Union[str, pandas.DataFrame, h2o.DataFrame, pyspark.sql.dataframe.DataFrame]
            If str, URI for the dataset. The dataset could be read from local or network file system, hdfs, s3, gcs and optionally pyspark in pyspark
            conda env
        target: str, optional
            Name of the target in dataset.
            If set an ADSDatasetWithTarget object is returned, otherwise an ADSDataset object is returned which can be
            used to understand the dataset through visualizations
        format: str, default: infer
            Format of the dataset.
            Supported formats: CSV, TSV, Parquet, libsvm, JSON, XLS/XLSX (Excel), HDF5, SQL, XML,
            Apache server log files (clf, log), ARFF.
            By default, the format would be inferred from the ending of the dataset file path.
        reader_fn: Callable, default: None
            The user may pass in their own custom reader function.
            It must accept `(path, **kwarg)` and return a pandas DataFrame
        name: str, optional default: ""
        description: str, optional default: ""
            Text describing the dataset
        npartitions: int, deprecated
            Number of partitions to split the data
            By default this is set to the max number of cores supported by the backend compute accelerator
        type_discovery: bool, default: True
            If false, the data types of the dataframe are used as such.
            By default, the dataframe columns are associated with the best suited data types. Associating the features
            with the disovered datatypes would impact visualizations and model prediction.
        html_table_index: int, optional
            The index of the dataframe table in html content. This is used when the format of dataset is html
        column_names: 'infer', list of str or None, default: 'infer'
            Supported only for CSV and TSV.
            List of column names to use.
            By default, column names are inferred from the first line of the file.
            If set to None, column names would be auto-generated instead of inferring from file.
            If the file already contains a column header, specify header=0 to ignore the existing column names.
        sample_max_rows: int, default: 10000, use -1 auto calculate sample size, use 0 (zero) for no sampling
            Sample size of the dataframe to use for visualization and optimization.
        positive_class: Any, optional
            Label in target for binary classification problems which should be identified as positive for modeling.
            By default, the first unique value is considered as the positive label.
        types: dict, optional
            Dictionary of <feature_name> : <data_type> to override the data type of features.
        transformer_pipeline: datasets.pipeline.TransformerPipeline, optional
            A pipeline of transformations done outside the sdk and need to be applied at the time of scoring
        storage_options: dict, default: varies by source type
            Parameters passed on to the backend filesystem class.
        sep: str
            Delimiting character for parsing the input file.
        kwargs: additional keyword arguments that would be passed to underlying dataframe read API
            based on the format of the dataset

        Returns
        -------
        dataset : An instance of ADSDataset
        (or)
        dataset_with_target : An instance of ADSDatasetWithTarget

        Examples
        --------
        >>> ds = DatasetFactory.open("/path/to/data.data", format='csv', delimiter=" ",
        ...          na_values="n/a", skipinitialspace=True)

        >>> ds = DatasetFactory.open("/path/to/data.csv", target="col_1", prefix="col_",
        ...           skiprows=1, encoding="ISO-8859-1")

        >>> ds = DatasetFactory.open("oci://bucket@namespace/path/to/data.tsv",
        ...         column_names=["col1", "col2", "col3"], header=0)

        >>> ds = DatasetFactory.open("oci://bucket@namespace/path/to/data.csv",
        ...         storage_options={"config": "~/.oci/config",
        ...         "profile": "USER_2"}, delimiter = ';')

        >>> ds = DatasetFactory.open("/path/to/data.parquet", engine='pyarrow',
        ...         types={"col1": "ordinal",
        ...                "col2": "categorical",
        ...                "col3" : "continuous",
        ...                "col4" : "float64"})

        >>> ds = DatasetFactory.open(df, target="class", sample_max_rows=5000,
        ...          positive_class="yes")

        >>> ds = DatasetFactory.open("s3://path/to/data.json.gz", format="json",
        ...         compression="gzip", orient="records")
        """
        if npartitions:
            warnings.warn(
                "Variable `npartitions` is deprecated and will not be used",
                DeprecationWarning,
                stacklevel=2,
            )
        if (
            "storage_options" not in kwargs
            and type(source) is str
            and len(source) > 6
            and source[:6] == "oci://"
        ):
            kwargs["storage_options"] = {"config": {}}

        if isinstance(source, str) or isinstance(source, list):
            progress = utils.get_progress_bar(4)
            progress.update("Opening data")
            path = ElaboratedPath(source, format=format, **kwargs)
            reader_fn = (
                get_format_reader(path=path, **kwargs)
                if reader_fn is None
                else reader_fn
            )
            df = load_dataset(path=path, reader_fn=reader_fn, **kwargs)
            name = path.name
        elif isinstance(source, pd.DataFrame):
            progress = utils.get_progress_bar(4)
            progress.update("Partitioning data")
            df = source
            name = "User Provided DataFrame" if name is None else name
        else:
            raise TypeError(
                f"The Source type: {type(source)} is not supported for DatasetFactory."
            )
        shape = df.shape
        return DatasetFactory._build_dataset(
            df=df,
            shape=shape,
            target=target,
            sample_max_rows=sample_max_rows,
            type_discovery=type_discovery,
            types=types,
            positive_class=positive_class,
            name=name,
            transformer_pipeline=transformer_pipeline,
            description=description,
            progress=progress,
            **utils.inject_and_copy_kwargs(
                kwargs,
                **{"html_table_index": html_table_index, "column_names": column_names},
            ),
        )

    @staticmethod
    def open_to_pandas(
        source: str, format: str = None, reader_fn: Callable = None, **kwargs
    ) -> pd.DataFrame:
        path = ElaboratedPath(source, format=format, **kwargs)
        reader_fn = (
            get_format_reader(path=path, **kwargs) if reader_fn is None else reader_fn
        )
        df = load_dataset(path=path, reader_fn=reader_fn, **kwargs)
        return df

    @staticmethod
    def from_dataframe(df, target: str = None, **kwargs):
        """
        Returns an object of ADSDatasetWithTarget or ADSDataset given a pandas.DataFrame

        Parameters
        ----------
        df: pandas.DataFrame
        target: str
        kwargs: dict
            See DatasetFactory.open() for supported kwargs

        Returns
        -------
        dataset: an object of ADSDataset target is not specified, otherwise an object of ADSDatasetWithTarget tagged
                  according to the type of target

        Examples
        --------
        >>> df = pd.DataFrame(data)
        >>> ds = from_dataframe(df)
        """
        return DatasetFactory.open(df, target=target, **kwargs)

    @staticmethod
    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    @runtime_dependency(
        module="ipywidgets",
        object="HTML",
        is_for_notebook_only=True,
        install_from=OptionalDependency.NOTEBOOK,
    )
    def list_snapshots(snapshot_dir=None, name="", storage_options=None, **kwargs):
        """
        Displays the URIs for dataset snapshots under the given directory path.

        Parameters
        ----------
        snapshot_dir: str
            Return all dataset snapshots created using ADSDataset.snapshot() within this directory.
            The path can contain protocols such as oci, s3.
        name: str, optional
            The list of snapshots in the directory gets filtered by the name. Accepts glob expressions.
            default = `"ads_"`
        storage_options: dict
            Parameters passed on to the backend filesystem class.

        Example
        --------
        >>> DatasetFactory.list_snapshots(snapshot_dir="oci://my_bucket/snapshots_dir",
        ...             name="ads_iris_")

        Returns a list of all snapshots (recursively) saved to obj storage bucket `"my_bucket"` with prefix
        `"/snapshots_dir/ads_iris_**"` sorted by time created.
        """
        if snapshot_dir is None:
            snapshot_dir = default_snapshots_dir
            if snapshot_dir is None:
                raise ValueError(
                    "Specify snapshot_dir or use DatasetFactory.set_default_storage() to set default \
                                      storage options"
                )
            else:
                logger.info("Using default snapshots dir %s" % snapshot_dir)
        if storage_options is None:
            if default_storage_options is not None:
                storage_options = default_storage_options
                logger.info("Using default storage options")
            else:
                storage_options = dict()
        assert isinstance(storage_options, dict), (
            "The storage options parameter must be a dictionary. You can set "
            "this gloabally by calling DatasetFactory.set_default_storage("
            "storage_options={'config': 'location'}). "
        )
        url_options = infer_storage_options(snapshot_dir)
        protocol = url_options.pop("protocol", None)

        fs = OCIFileSystem(config=storage_options.get("config", None))
        kwargs.update({"refresh": True})
        obj_list = [
            (k, v.get("timeCreated", mindate).strftime("%Y-%m-%d %H:%M:%S"))
            for k, v in fs.glob(
                os.path.join(snapshot_dir, name + "**"), detail=True, **kwargs
            ).items()
            if v["type"] == "file"
        ]

        files = []
        for file, file_time in obj_list:
            if protocol in ["oci"]:
                r1 = re.compile(r"/part\.[0-9]{1,6}\.parquet$")
                parquet_part = r1.search(file)
                if parquet_part is not None:
                    parquet_filename = file[: parquet_part.start()]
                elif file.endswith("/_common_metadata"):
                    parquet_filename = file[: -len("/_common_metadata")]
                elif file.endswith("/_metadata"):
                    parquet_filename = file[: -len("/_metadata")]
                else:
                    parquet_filename = file
            else:
                parquet_filename = file
            parent_path = "%s://" % protocol
            files.append((parent_path + parquet_filename, file_time))
        files.sort(key=lambda x: x[1] or mindate, reverse=True)
        list_df = pd.DataFrame(files, columns=["Name", "Created Time"])
        list_df = list_df.drop_duplicates(subset=["Name"]).reset_index()
        if len(list_df) == 0:
            print(f"No snapshots found at: {os.path.join(snapshot_dir, name)}")

        # display in HTML format if sdk is run in notebook mode
        if utils.is_notebook():
            from IPython.core.display import display

            display(
                HTML(
                    list_df.style.set_table_attributes("class=table")
                    .hide_index()
                    .render()
                )
            )
        return list_df

    @staticmethod
    def download(remote_path, local_path, storage=None, overwrite=False):
        """
        Download a remote file or directory to local storage.

        Parameters
        ---------
        remote_path: str
            Supports protocols like oci, s3, also supports glob expressions
        local_path: str
            Supports glob expressions
        storage: dict
            Parameters passed on to the  backend remote filesystem class.
        overwrite: bool, default False
            If True, the method will overwrite any existing files in the local_path

        Examples
        ---------
        >>> DatasetFactory.download("oci://Bucket/prefix/to/data/*.csv",
        ...         "/home/datascience/data/")
        """
        if storage is None:
            if default_storage_options is not None:
                storage = default_storage_options
                logger.info("Using default storage options")
            else:
                storage = dict()

        remote_files = fsspec.open_files(
            remote_path, mode="rb", name_function=lambda i: "", **storage
        )
        if len(remote_files) < 1:
            raise FileNotFoundError(remote_path)
        display_error, error_msg = DatasetFactory._download_files(
            remote_files=remote_files, local_path=local_path, overwrite=overwrite
        )
        if display_error:
            logger.error(error_msg)
        else:
            logger.info(f"Download {remote_path} to {local_path}.")

    @staticmethod
    def _download_files(remote_files, local_path, overwrite=False):
        display_error, error_msg = False, ""
        for remote_file in remote_files:
            bucket_idx = remote_file.path.find("/")
            suffix = remote_file.path[bucket_idx + 1 :]
            try:
                with remote_file as f1:
                    local_filepath = (
                        os.path.join(local_path, suffix) if suffix else local_path
                    )
                    if os.path.exists(local_filepath) and not overwrite:
                        raise FileExistsError(
                            f"Trying to overwrite files in {local_filepath}. If you'd like to "
                            f"overwrite these files, set force_overwrite to True."
                        )
                    os.makedirs(os.path.dirname(local_filepath), exist_ok=True)
                    with open(local_filepath, "wb") as f2:
                        f2.write(f1.read())
            except oci.exceptions.ServiceError as e:
                raise FileNotFoundError(f"Unable to open file: {remote_file.path}")
        return display_error, error_msg

    @staticmethod
    def upload(local_file_or_dir, remote_file_or_dir, storage_options=None):
        """
        Upload local file or directory to remote storage

        Parameters
        ---------
        local_file_or_dir: str
            Supports glob expressions
        remote_file_or_dir: str
            Supports protocols like oci, s3, also supports glob expressions
        storage_options: dict
            Parameters passed on to the  backend remote filesystem class.
        """
        if not os.path.exists(local_file_or_dir):
            raise ValueError("File/Directory does not exist: %s" % local_file_or_dir)
        if storage_options is None and default_storage_options is not None:
            storage_options = default_storage_options
            logger.info("Using default storage options")

        if os.path.isdir(local_file_or_dir):
            for subdir, dirs, files in os.walk(local_file_or_dir):
                for file in files:
                    if os.path.abspath(subdir) == os.path.abspath(local_file_or_dir):
                        path = file
                    else:
                        path = os.path.join(
                            os.path.abspath(subdir).split("/", 2)[2], file
                        )
                    DatasetFactory._upload_file(
                        os.path.join(subdir, file),
                        os.path.join(remote_file_or_dir, path),
                        storage_options=storage_options,
                    )
        else:
            DatasetFactory._upload_file(
                local_file_or_dir, remote_file_or_dir, storage_options=storage_options
            )

    @staticmethod
    def set_default_storage(snapshots_dir=None, storage_options=None):
        """
        Set default storage directory and options.

        Both snapshots_dir and storage_options can be overridden at the API scope.

        Parameters
        ----------
        snapshots_dir: str
            Path for the snapshots directory. Can contain protocols such as oci, s3
        storage_options: dict, optional
             Parameters passed on to the backend filesystem class.
        """
        global default_snapshots_dir
        default_snapshots_dir = snapshots_dir
        global default_storage_options
        if storage_options is not None:
            assert isinstance(storage_options, dict), (
                f"The storage options parameter must be a dictionary. Instead "
                f"we got the type: {type(storage_options)} "
            )
        default_storage_options = storage_options

    @classmethod
    def _upload_file(cls, local_file, remote_file, storage_options=None):
        kwargs = {}
        if storage_options is not None:
            kwargs = {"storage_options": storage_options}
        remote_file_handler = fsspec.open_files(
            remote_file + "*", mode="wb", name_function=lambda i: "", **kwargs
        )[0]
        with remote_file_handler as f1:
            with open(local_file, "rb") as f2:
                for line in f2:
                    f1.write(line)
        print("Uploaded %s to %s" % (local_file, remote_file))

    @classmethod
    def _build_dataset(
        cls,
        df: pd.DataFrame,
        shape: Tuple[int, int],
        target: str = None,
        progress=None,
        **kwargs,
    ):
        n = shape[0]
        if progress:
            progress.update("Generating data sample")

        sampled_df = generate_sample(
            df,
            n,
            DatasetDefaults.sampling_confidence_level,
            DatasetDefaults.sampling_confidence_interval,
            **kwargs,
        )

        if target is None:
            if progress:
                progress.update("Building the dataset with no target.")
            result = ADSDataset(df=df, sampled_df=sampled_df, shape=shape, **kwargs)
            if progress:
                progress.update("Done")
            logger.info(
                "Use `set_target()` to type the dataset for a particular learning task."
            )
            return result

        if progress:
            progress.update("Building dataset")

        discover_target_type = kwargs["type_discovery"]
        if target in kwargs["types"]:
            sampled_df[target] = sampled_df[target].astype(kwargs["types"][target])
            discover_target_type = False

        # if type discovery is turned off, infer type from pandas dtype
        target_type = DatasetFactory.infer_target_type(
            target, sampled_df[target], discover_target_type
        )

        result = DatasetFactory._get_dataset(
            df=df,
            sampled_df=sampled_df,
            target=target,
            target_type=target_type,
            shape=shape,
            **kwargs,
        )
        if progress:
            progress.update("Done")
        logger.info(
            "Use `suggest_recommendations()` to view and apply recommendations for dataset optimization."
        )
        return result

    @classmethod
    def infer_target_type(cls, target, target_series, discover_target_type=True):
        # if type discovery is turned off, infer type from pandas dtype
        if discover_target_type:
            target_type = TypeDiscoveryDriver().discover(
                target, target_series, is_target=True
            )
        else:
            target_type = get_feature_type(target, target_series)
        return target_type

    @classmethod
    def _get_dataset(
        cls,
        df: pd.DataFrame,
        sampled_df: pd.DataFrame,
        target: str,
        target_type: TypedFeature,
        shape: Tuple[int, int],
        positive_class=None,
        **init_kwargs,
    ):
        if len(df[target].dropna()) == 0:
            logger.warning(
                "It is not recommended to use an empty column as the target variable."
            )
            raise ValueError(
                f"We do not support using empty columns as the chosen target"
            )
        if is_same_class(target_type, ContinuousTypedFeature):
            return RegressionDataset(
                df=df,
                sampled_df=sampled_df,
                target=target,
                target_type=target_type,
                shape=shape,
                **init_kwargs,
            )
        elif is_same_class(
            target_type, DateTimeTypedFeature
        ) or df.index.dtype.name.startswith("datetime"):
            return ForecastingDataset(
                df=df,
                sampled_df=sampled_df,
                target=target,
                target_type=target_type,
                shape=shape,
                **init_kwargs,
            )

        # Adding ordinal typed feature, but ultimately we should rethink how we want to model this type
        elif is_same_class(target_type, CategoricalTypedFeature) or is_same_class(
            target_type, OrdinalTypedFeature
        ):
            if target_type.meta_data["internal"]["unique"] == 2:
                if is_text_data(sampled_df, target):
                    return BinaryTextClassificationDataset(
                        df=df,
                        sampled_df=sampled_df,
                        target=target,
                        shape=shape,
                        target_type=target_type,
                        positive_class=positive_class,
                        **init_kwargs,
                    )

                return BinaryClassificationDataset(
                    df=df,
                    sampled_df=sampled_df,
                    target=target,
                    shape=shape,
                    target_type=target_type,
                    positive_class=positive_class,
                    **init_kwargs,
                )
            else:
                if is_text_data(sampled_df, target):
                    return MultiClassTextClassificationDataset(
                        df=df,
                        sampled_df=sampled_df,
                        target=target,
                        target_type=target_type,
                        shape=shape,
                        **init_kwargs,
                    )
                return MultiClassClassificationDataset(
                    df=df,
                    sampled_df=sampled_df,
                    target=target,
                    target_type=target_type,
                    shape=shape,
                    **init_kwargs,
                )
        elif (
            is_same_class(target, DocumentTypedFeature)
            or "text" in target_type["type"]
            or "text" in target
        ):
            raise ValueError(
                f"The column {target} cannot be used as the target column."
            )
        elif (
            is_same_class(target_type, GISTypedFeature)
            or "coord" in target_type["type"]
            or "coord" in target
        ):
            raise ValueError(
                f"The column {target} cannot be used as the target column."
            )
        # This is to catch constant columns that are boolean. Added as a fix for pd.isnull(), and datasets with a
        #   binary target, but only data on one instance
        elif target_type["low_level_type"] == "bool":
            return BinaryClassificationDataset(
                df=df,
                sampled_df=sampled_df,
                target=target,
                shape=shape,
                target_type=target_type,
                positive_class=positive_class,
                **init_kwargs,
            )
        raise ValueError(
            f"Unable to identify problem type. Specify the data type of {target} using 'types'. "
            f"For example, types = {{{target}: 'category'}}"
        )


class CustomFormatReaders:
    @staticmethod
    def read_tsv(path: str, **kwargs) -> pd.DataFrame:
        return pd.read_csv(
            path, **utils.inject_and_copy_kwargs(kwargs, **{"sep": "\t"})
        )

    @staticmethod
    def read_json(path: str, **kwargs) -> pd.DataFrame:
        try:
            return pd.read_json(path, **kwargs)
        except ValueError as e:
            return pd.read_json(
                path, **utils.inject_and_copy_kwargs(kwargs, **{"lines": True})
            )

    @staticmethod
    def read_libsvm(path: str, **kwargs) -> pd.DataFrame:
        from sklearn.datasets import load_svmlight_file
        from joblib import Memory

        mem = Memory("./mycache")

        @mem.cache
        def get_data(path):
            X, y = load_svmlight_file(path)
            df = pd.DataFrame(X.todense())
            df["target"] = y
            return df

        return get_data(path)

    @staticmethod
    @runtime_dependency(
        module="pandavro", object="read_avro", install_from=OptionalDependency.DATA
    )
    def read_avro(path: str, **kwargs) -> pd.DataFrame:
        return read_avro(path, **kwargs)

    DEFAULT_SQL_CHUNKSIZE = 12007
    DEFAULT_SQL_ARRAYSIZE = 50000
    DEFAULT_SQL_MIL = 128
    DEFAULT_SQL_CTU = False

    @classmethod
    def read_sql(cls, path: str, table: str = None, **kwargs) -> pd.DataFrame:
        """

        :param path: str
            This is the connection URL that gets passed to sqlalchemy's create_engine method
        :param table: str
            This is either the name of a table to select * from or a sql query to be run
        :param kwargs:
        :return: pd.DataFrame
        """
        if table is None:
            raise ValueError(
                "In order to read from a database you need to specify the table using the `table` "
                "argument."
            )
        # check if it's oracle dialect
        if str(path).lower().startswith("oracle"):
            kwargs = utils.inject_and_copy_kwargs(
                kwargs,
                **{
                    "arraysize": cls.DEFAULT_SQL_ARRAYSIZE,
                    "max_identifier_length": cls.DEFAULT_SQL_MIL,
                    "coerce_to_unicode": cls.DEFAULT_SQL_CTU,
                },
            )
        engine = utils.get_sqlalchemy_engine(path, **kwargs)

        table_name = table.strip()
        with engine.connect() as connection:
            # if it's a query expression:
            if table_name.lower().startswith("select"):
                sql_query = table_name
            else:
                sql_query = f"select * from {table_name}"

            chunks = pd.read_sql_query(
                sql_query,
                con=connection,
                **_validate_kwargs(
                    pd.read_sql_query,
                    utils.inject_and_copy_kwargs(
                        kwargs, **{"chunksize": cls.DEFAULT_SQL_CHUNKSIZE}
                    ),
                ),
            )
            df = pd.DataFrame()
            from tqdm import tqdm

            with tqdm(chunks, unit=" rows") as t:
                for chunk in chunks:
                    df = pd.concat([df, chunk])
                    t.update(len(chunk))

            df = df.reset_index(drop=True)
            if df.shape[0] == 0:
                logger.warning(
                    "The SQL expression returned zero rows. Therefore, no `ADSdataset` object was created."
                )
                raise Exception("The SQL expression returned no rows")
        return df

    @staticmethod
    def read_log(path, **kwargs):
        from ads.dataset.helper import parse_apache_log_str, parse_apache_log_datetime

        df = pd.read_csv(
            path,
            # assume_missing=True,
            sep=r'\s(?=(?:[^"]*"[^"]*")*[^"]*$)(?![^\[]*\])',
            engine="python",
            na_values="-",
            header=None,
            names=[
                "host",
                "identity",
                "user",
                "time",
                "request",
                "http_code",
                "response_bytes",
                "referer",
                "user_agent",
                "unknown",
            ],
            converters={
                "time": parse_apache_log_datetime,
                "request": parse_apache_log_str,
                "status": int,
                "size": int,
                "referer": parse_apache_log_str,
                "user_agent": parse_apache_log_str,
            },
            **kwargs,
        )
        return df

    @staticmethod
    def read_html(path, html_table_index: int = None, **kwargs):
        if html_table_index is None:
            return pd.concat(df for df in pd.read_html(path, **kwargs))
        else:
            return pd.read_html(path, **kwargs)[html_table_index]

    @staticmethod
    @runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
    def read_arff(path, **kwargs):
        from scipy.io import arff
        import requests
        from io import BytesIO, TextIOWrapper

        data = None
        if os.path.isfile(path):
            data, _ = arff.loadarff(path)
        else:
            with requests.get(path) as r:
                if r.status_code == requests.codes.ok:
                    f = TextIOWrapper(BytesIO(r.content))
                    data, _ = arff.loadarff(f)
        return pd.DataFrame(data)

    @staticmethod
    def read_xml(path: str, **kwargs) -> pd.DataFrame:
        """
        Load data from xml file.

        Parameters
        ----------
        path: str
            Path to XML file
        storage_options: dict, optional
            Storage options passed to Pandas to read the file.

        Returns
        -------
        dataframe : pandas.DataFrame
        """
        import xml.etree.cElementTree as et

        def get_children(df, node, parent, i):

            for name in node.attrib.keys():
                df.at[i, parent + name] = node.attrib[name]
            for child in list(node):
                if len(list(child)) > 0:
                    get_children(df, child, parent + child.tag + "/", i)
                else:
                    df.at[i, parent + child.tag] = child.text

        storage_options = kwargs.get("storage_options", {})

        file_handles = fsspec.open_files(path, mode="rb", **storage_options)
        ret_df = pd.DataFrame()
        last_i = 0
        for file_handle in file_handles:
            with file_handle:
                parsed_xml = et.parse(path)
                for i, node in enumerate(parsed_xml.getroot()):
                    get_children(ret_df, node, node.tag + "/", last_i + i)
                last_i = i
        return ret_df


reader_fns = {
    "csv": pd.read_csv,
    "tsv": CustomFormatReaders.read_tsv,
    "json": CustomFormatReaders.read_json,
    "jsonl": CustomFormatReaders.read_json,
    "excel": pd.read_excel,
    "xls": pd.read_excel,
    "xlsx": pd.read_excel,
    "parquet": pd.read_parquet,
    "libsvm": CustomFormatReaders.read_libsvm,
    "hdf": pd.read_hdf,  # Todo: re.match(format, "hdf\d*") or format == "h5"
    "hdf3": pd.read_hdf,
    "hdf4": pd.read_hdf,
    "h5": pd.read_hdf,
    "avro": CustomFormatReaders.read_avro,
    "avsc": CustomFormatReaders.read_avro,
    "sql": CustomFormatReaders.read_sql,
    "db": CustomFormatReaders.read_sql,
    "log": CustomFormatReaders.read_log,
    "clf": CustomFormatReaders.read_log,
    "html": CustomFormatReaders.read_html,
    "arff": CustomFormatReaders.read_arff,
    "xml": CustomFormatReaders.read_xml,
}


def _validate_kwargs(func: Callable, kwargs):
    valid_params = inspect.signature(func).parameters
    if "kwargs" in valid_params:
        return kwargs
    else:
        return {k: v for k, v in kwargs.items() if k in valid_params}


def get_format_reader(path: ElaboratedPath, **kwargs) -> Callable:
    format_key = path.format
    try:
        reader_fn = reader_fns[format_key]
    except (KeyError, NameError):
        raise ValueError(
            f"We were unable to load the specified dataset. We have interpreted the format "
            f"as {format_key}, if this is not correct, call again and set the `format` parameter = "
            f"to the desired format. Read more here: https://docs.cloud.oracle.com/en-us/iaas/tools/ads"
            f"-sdk/latest/user_guide/loading_data/loading_data.html#specify-data-types-in-load-dataset"
        )

    return reader_fn


def load_dataset(path: ElaboratedPath, reader_fn: Callable, **kwargs) -> pd.DataFrame:
    dfs = []
    for filename in path.paths:
        data = reader_fn(filename, **_validate_kwargs(reader_fn, kwargs))
        if not isinstance(data, pd.DataFrame):
            fn_name = f"{reader_fn.__module__}.{reader_fn.__qualname__}"
            raise ValueError(
                f"{fn_name} is used to load the data. "
                f"However, {fn_name} returned {type(data)} instead of pandas DataFrame. "
                f"Refer to the usage of {fn_name} to set the correct arguments."
            )
        dfs.append(data)
    if len(dfs) == 0:
        raise ValueError(
            f"We were unable to load the specified dataset. Read more here: "
            f"https://docs.cloud.oracle.com/en-us/iaas/tools/ads"
            f"-sdk/latest/user_guide/loading_data/loading_data.html#specify-data-types-in-load-dataset"
        )

    df = pd.concat(dfs)

    if df is None:
        raise ValueError(
            f"We were unable to load the specified dataset. Read more here: "
            f"https://docs.cloud.oracle.com/en-us/iaas/tools/ads"
            f"-sdk/latest/user_guide/loading_data/loading_data.html#specify-data-types-in-load-dataset"
        )
    if df.empty:
        raise DatasetLoadException("Empty DataFrame, not producing a ADSDataset")
    return df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""
This is a driver script from Oracle ADS to run Python script in OCI Data Science Jobs.
The following environment variables are used:
GIT_URL:
    URL to the Git repository.
GIT_BRANCH:
    Optional, the Git branch to checkout.
GIT_COMMIT:
    Optional, the Git commit to checkout. By default, the most recent git commit will be checked out.
CODE_DIR:
    Optional, the directory for saving the user code from Git repository. Defaults to "~/Code"
GIT_ENTRYPOINT:
    Relative path to the entry script/module file in the Git repository.
ENTRY_FUNCTION:
    Optional, function name in the entry script/module to be invoked.
    If this is not specified, the entry script will be run as Python script.
PYTHON_PATH:
    Optional, additional paths to be added to sys.path for looking up modules and packages.
    The root of the Git repository will be added by default.
    Multiple paths can be separated by os.pathsep, which is colon(:) for Linux and Mac, semicolon(;) for Windows.
OUTPUT_DIR:
    Optional, output directory to be copied to object storage.
OUTPUT_URI:
    Optional, object storage URI for saving files from the output directory.
GIT_SECRET_OCID:
    The OCID of the OCI vault secret storing the SSH key for Git commands.
SKIP_METADATA_UPDATE:
    If this variable exists, the update metadata step will be skipped.
OCI_IAM_TYPE:
    Authentication method for OCI services.
    OCI API key will be used if this is set to api_key.
    Otherwise resource principal will be used.
OCI_CONFIG_LOCATION:
    The location of OCI API key when OCI_IAM_TYPE is set to api_key.
    If this is not set, oci.config.DEFAULT_LOCATION will be used.
OCI_CONFIG_PROFILE:
    The profile name to be used for API key authentication.
    If this is not set, oci.config.DEFAULT_PROFILE will be used.
OCI__GIT_SSH_KEY_PATH:
    The location to save the SSH Key for accessing the git repository.

JOB_RUN_OCID:
    The OCID of the job run. This is set by the job run.

This module requires the following packages:
oci
requests
GitPython
git
openssh

"""
import base64
import logging
import os
import random
import shutil
import string
import traceback
import uuid
from time import sleep, time
from typing import Optional
from urllib.parse import urlparse
from urllib.request import getproxies

import git
import oci
import requests
from oci.data_science import DataScienceClient

try:
    # This is used by ADS and testing
    from .driver_utils import (
        OCIHelper,
        JobRunner,
        set_log_level,
        CONST_ENV_ENTRY_FUNC,
        CONST_ENV_CODE_DIR,
        CONST_ENV_JOB_RUN_OCID,
    )
except ImportError:
    # This is used when the script is in a job run.
    from driver_utils import (
        OCIHelper,
        JobRunner,
        set_log_level,
        CONST_ENV_ENTRY_FUNC,
        CONST_ENV_CODE_DIR,
        CONST_ENV_JOB_RUN_OCID,
    )


CONST_ENV_ENTRYPOINT = "GIT_ENTRYPOINT"
CONST_ENV_GIT_URL = "GIT_URL"
CONST_ENV_GIT_BRANCH = "GIT_BRANCH"
CONST_ENV_GIT_COMMIT = "GIT_COMMIT"
CONST_ENV_SECRET_OCID = "GIT_SECRET_OCID"

DEFAULT_CODE_DIR = "~/Code/"
SSH_DIR = os.path.join("/home", "datascience", "ssh_" + str(uuid.uuid4()))
SSH_KEY_FILE_PATH = os.path.join(SSH_DIR, "id_rsa")
SSH_CONFIG_FILE_PATH = os.path.join(SSH_DIR, "config")


logger = logging.getLogger(__name__)
logger = set_log_level(logger)


class GitManager:
    """Contains methods for fetching code from Git repository"""

    def __init__(
        self, repo_url: str, code_dir: str = os.environ.get(CONST_ENV_CODE_DIR)
    ):
        """Initialize the GitManager

        Parameters
        ----------
        repo_url : str
            The URL of the repository.
        code_dir : str
            The local directory for storing the code from Git repository.

        Raises
        ------
        ValueError
            URL is not specified.
        """
        if not repo_url:
            raise ValueError("Specify the URL of Git repository.")
        self.repo_url = repo_url
        # Use default directory if code_dir is set to None or empty string.
        if not code_dir:
            code_dir = os.path.join(
                DEFAULT_CODE_DIR, os.path.basename(repo_url).split(".", 1)[0]
            )
        self.code_dir = os.path.abspath(os.path.expanduser(code_dir))

        logger.info("Initializing code directory at %s", self.code_dir)
        # Rename the existing directory if one already exists
        if os.path.exists(self.code_dir) and os.listdir(self.code_dir):
            logger.warning(
                "Directory %s already exists and is not empty.", self.code_dir
            )
            new_name = (
                self.code_dir
                + "_"
                + "".join(random.choice(string.ascii_lowercase) for i in range(5))
            )
            shutil.move(
                self.code_dir,
                new_name,
            )
            logger.warning("Renamed %s to %s", self.code_dir, new_name)
        os.makedirs(self.code_dir, exist_ok=True)

        self.repo = None
        self.commit = None

    def _config_ssh_proxy(self, proxy):
        return_code = JobRunner.run_command("command -v socat", level=logging.DEBUG)
        if return_code:
            logger.warning(
                "You have ssh_proxy configured. "
                "Please install the 'socat' package into your environment "
                "if you would like to use proxy for Git clone via SSH."
            )
        else:
            ssh_config = f"ProxyCommand socat - PROXY:{proxy.hostname}:%h:%p,proxyport={proxy.port}"
            logger.debug("Adding proxy for SSH: %s", ssh_config)
            with open(SSH_CONFIG_FILE_PATH, "a", encoding="utf-8") as f:
                f.write(ssh_config)
            logger.debug("SSH config saved to %s", SSH_CONFIG_FILE_PATH)

    def _config_known_hosts(self, host: str):
        if JobRunner.run_command(
            f"ssh-keyscan -H {host} >> {SSH_DIR}/known_hosts", level=logging.DEBUG
        ):
            logger.debug("Added %s to known hosts.", host)
        else:
            logger.warning(
                "Failed to add %s to known hosts."
                "You may need to configure your subnet security list to allow traffic on port 22.",
                host,
            )
        # Test the connection, for debugging purpose
        if logger.level == logging.DEBUG:
            os.system(f"ssh -vT git@{host}")

    def _config_ssh_key(self):
        if os.path.exists(SSH_KEY_FILE_PATH):
            # Ignore the fingerprint checking
            ssh_cmd = f'ssh -i "{SSH_KEY_FILE_PATH}" '
            if os.path.exists(SSH_CONFIG_FILE_PATH):
                ssh_cmd += f'-F "{SSH_CONFIG_FILE_PATH}" '
            ssh_cmd += (
                '-o "StrictHostKeyChecking=no" -o "UserKnownHostsFile=/dev/null" '
            )
            if logger.level == logging.DEBUG:
                ssh_cmd += "-v "
            logger.debug("SSH command: %s", ssh_cmd)
            return {"GIT_SSH_COMMAND": ssh_cmd}
        return None

    def fetch_repo(self):
        """Clones the Git repository."""
        logger.debug("Cloning from %s", self.repo_url)
        env = None
        if self.repo_url.startswith(("git@", "ssh://")):
            proxies = getproxies()
            if ("http" in proxies or "https" in proxies) and "ssh" not in proxies:
                logger.warning(
                    "You have http/https proxy configured. "
                    "Please set ssh_proxy in environment variable and install the 'socat' package "
                    "if you would like to use the proxy for Git clone via SSH. "
                    "conda install -c conda-forge socat"
                )
            if "ssh" in proxies:
                proxy = urlparse(proxies["ssh"])
                self._config_ssh_proxy(proxy=proxy)
            env = self._config_ssh_key()

        if logger.level == logging.DEBUG:
            os.system("git --version")

        self.repo = git.Repo.clone_from(self.repo_url, self.code_dir, env=env)
        logger.info("Cloned repo to: %s", self.code_dir)
        return self

    def checkout_code(self, branch: Optional[str] = None, commit: Optional[str] = None):
        """Checkouts the branch or commit of the Git repository.

        If neither branch nor commit is specified, the tip of the default branch will be used.
        If both branch and commit are specified, the commit will be used.

        Parameters
        ----------
        branch : str, optional
            The name of the branch, by default None
        commit : str, optional
            The commit ID (SHA1 hash), by default None

        """
        if commit:
            logger.info("Checking out commit: %s", commit)
            self.repo.git.checkout(commit)
        elif branch:
            logger.info("Checking out branch: %s", branch)
            self.repo.git.checkout(branch)
        else:
            logger.info(
                "Checking out the latest commit %s", self.repo.head.commit.hexsha
            )
            self.commit = self.repo.head.commit.hexsha
        return self


class CredentialManager:
    @staticmethod
    def read_secret(secret_id):
        """Reads and decode the value of of a secret from OCI vault.

        Parameters
        ----------
        secret_id : str
            OCID of the secret

        Returns
        -------
        str
            The value of the secret decoded with ASCII.
        """
        secret_client = OCIHelper.init_oci_client(oci.secrets.SecretsClient)
        secret_bundle = secret_client.get_secret_bundle(secret_id)
        base64_secret_bytes = secret_bundle.data.secret_bundle_content.content.encode(
            "ascii"
        )
        base64_message_bytes = base64.b64decode(base64_secret_bytes)
        secret_content = base64_message_bytes.decode("ascii")
        return secret_content


class GitSSHKey:
    def __init__(self, secret_id) -> None:
        self.secret_id = secret_id
        self.key_file_path = os.path.expanduser(
            os.environ.get("OCI__GIT_SSH_KEY_PATH", SSH_KEY_FILE_PATH)
        )
        self.existing_git_ssh_cmd = None
        self.backup_file_path = None

    def _set_ssh_key(self):
        """Setup SSH key for Git command"""
        content = CredentialManager().read_secret(self.secret_id)
        # Add a new line to the SSH key in case the user forget to do so
        # SSH key without a new line at the end is considered "invalid format".
        if not content.endswith("\n"):
            content += "\n"
        os.makedirs(os.path.dirname(self.key_file_path), exist_ok=True)
        with open(self.key_file_path, "w", encoding="utf-8") as f:
            f.write(content)
        # Set the correct permission for the SSH key.
        os.chmod(self.key_file_path, 0o600)
        logger.info("SSH key saved to %s", self.key_file_path)

    def _backup_ssh_key(self):
        """Backup the existing SSH key if one exists at the same file location."""
        if not os.path.exists(self.key_file_path):
            return
        suffix = (
            "."
            + "".join(random.choice(string.ascii_lowercase) for i in range(10))
            + ".bak"
        )
        self.backup_file_path = self.key_file_path + suffix
        os.rename(self.key_file_path, self.backup_file_path)

    def _restore_backup(self):
        """Restore the SSH key from backup."""
        if self.backup_file_path and os.path.exists(self.backup_file_path):
            os.rename(self.backup_file_path, self.key_file_path)

    def __enter__(self):
        """Setup the SSH key for Git command"""
        if self.secret_id:
            self._backup_ssh_key()
            self._set_ssh_key()

    def __exit__(self, tp, value, tb):
        """Remove the SSH key and reset the git command option"""
        if not self.key_file_path:
            return
        # Remove the ssh key file only if there is secret
        if self.secret_id:
            try:
                os.remove(self.key_file_path)
                self._restore_backup()
            except OSError:
                # Since this will not affect any job outputs,
                # ignore the error if the file operation failed for some reason.
                pass


class GitJobRunner(JobRunner):
    """Contains methods for running the job."""

    def __init__(
        self,
        git_manager: GitManager,
        job_run_ocid: str = os.environ.get(CONST_ENV_JOB_RUN_OCID),
    ):
        """Initialize the job runner

        Parameters
        ----------
        job_run_ocid : str, optional
            Job run OCID, by default "".
        For local testing purpose, job run OCID can be set to empty string
        """
        self.job_run_ocid = job_run_ocid
        self.git_manager = git_manager
        self.artifacts = []
        self.source_info = {}
        super().__init__(code_dir=git_manager.code_dir)

    def run(
        self,
        entrypoint: str = os.environ.get(CONST_ENV_ENTRYPOINT),
        entry_function: str = os.environ.get(CONST_ENV_ENTRY_FUNC),
    ):
        """Runs the job

        Parameters
        ----------
        argv : list, optional
            A list of arguments for the entry script/function, by default None

        """
        self.source_info = {
            "repo": self.git_manager.repo_url,
            "commit": self.git_manager.commit,
            "module": entrypoint,
            "method": entry_function if entry_function else "",
        }

        entrypoint = os.path.join(self.code_dir, entrypoint)
        return super().run(entrypoint=entrypoint, entry_function=entry_function)

    @staticmethod
    def _raise_for_error(response: requests.Response):
        if response.status_code >= 400:
            try:
                message = response.json().get("message", "")
            except:
                message = response.content
            error = f"{response.status_code} {response.reason} Error for {response.url}\n{message}"
            raise requests.exceptions.HTTPError(error, response=response)

    def update_job_run_metadata_with_rest_api(
        self, client: DataScienceClient, metadata: dict
    ) -> None:
        """Updates the metadata of the job run by call OCI REST API.

        Parameters
        ----------
        client : DataScienceClient
            OCI DataScienceClient
        metadata : dict
            Metadata to be saved as freeform tags.
        """
        endpoint = f"{client.base_client.endpoint}/jobRuns/{self.job_run_ocid}"
        logger.debug("Request endpoint: %s", endpoint)
        headers = {"accept": "application/json", "content-type": "application/json"}
        signer = client.base_client.signer
        # Get the existing tags
        response = requests.get(endpoint, headers=headers, auth=signer)
        self._raise_for_error(response)

        tags = response.json().get("freeformTags", {})
        tags.update(metadata)
        body = {
            "definedTags": None,
            "displayName": None,
            "freeformTags": tags,
        }
        # Additional Headers for PUT request
        headers.update(
            {
                "expect": "100-continue",
                "opc-request-id": client.base_client.build_request_id(),
            }
        )
        response = requests.put(endpoint, headers=headers, json=body, auth=signer)
        self._raise_for_error(response)
        logger.debug(response.json())

    def save_metadata(self) -> None:
        """Saves the metadata to job run"""
        logger.info("Saving metadata to job run...")
        tags = {}
        # Source info
        for key, val in self.source_info.items():
            logger.debug("%s = %s", key, val)
            if val:
                tags[key] = str(val)
        # Output files
        if self.artifacts:
            prefix = "oci://" + os.path.commonprefix(self.artifacts)
            logger.debug("Job output: %s", prefix)
            tags["outputs"] = prefix

        client = OCIHelper.init_oci_client(DataScienceClient)
        oci.retry.DEFAULT_RETRY_STRATEGY.make_retrying_call(
            self.update_job_run_metadata_with_rest_api, client, tags
        )
        logger.debug("Updated Job Run metadata.")


def main():
    """The main function for running the job."""
    second_started = time()

    with GitSSHKey(os.environ.get(CONST_ENV_SECRET_OCID)):
        git_manager = (
            GitManager(os.environ.get(CONST_ENV_GIT_URL))
            .fetch_repo()
            .checkout_code(
                branch=os.environ.get(CONST_ENV_GIT_BRANCH, None),
                commit=os.environ.get(CONST_ENV_GIT_COMMIT, None),
            )
        )

    runner = GitJobRunner(git_manager)
    runner.set_working_dir().setup_python_path().run()

    # Copy outputs
    runner.artifacts = OCIHelper.copy_outputs()

    # Save metadata only if job run OCID is available
    if (
        os.environ.get(CONST_ENV_JOB_RUN_OCID)
        and "SKIP_METADATA_UPDATE" not in os.environ
    ):
        try:
            # Wait before updating the metadata.
            # Job run might still be in the ACCEPTED state shortly after it started,
            # Cannot update job run while in ACCEPTED, CANCELLING, DELETED or NEEDS_ATTENTION state
            second_elapsed = time() - second_started
            if second_elapsed < 90:
                sleep(90 - second_elapsed)
            runner.save_metadata()
        except Exception:
            logger.error("An error occurred when saving the metadata.")
            # Allow the job run to finish successfully even if the driver script failed to save the metadata.
            logger.debug(traceback.format_exc())

    logger.info("Job completed.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import operator
import types
import uuid
from functools import partial
from sklearn.preprocessing import FunctionTransformer
from ads.common.decorator.deprecate import deprecated

try:
    import dask
    from dask.dataframe.accessor import Accessor
    from dask.utils import OperatorMethodMixin
except ImportError as e:
    raise ModuleNotFoundError("Install dask to use the Dask Series class.") from e


class DaskSeries(dask.dataframe.core.Series, OperatorMethodMixin):
    """
    A DaskSeries object.
    """

    @deprecated("2.5.2")
    def __init__(self, series, key, transformers=[]):
        self.series = series
        self.key = key
        self.transformers = transformers

    def __getattribute__(self, item):
        return object.__getattribute__(self, "_getattribute")(item)

    def __getattr__(self, item):
        return object.__getattribute__(self, "_getattribute")(item)

    def _getattribute(self, item):
        try:
            attr = getattr(object.__getattribute__(self, "series"), item)
            if isinstance(attr, types.FunctionType) or isinstance(
                attr, types.MethodType
            ):
                return DaskSeries._apply(attr)
            elif isinstance(attr, Accessor):
                return DaskSeriesAccessor(attr, item)
            else:
                return attr
        except:
            attr = object.__getattribute__(self, item)
            return attr

    @classmethod
    def _get_unary_operator(cls, op):
        func = dask.dataframe.core.Series._get_unary_operator(op)
        return cls._apply(func)

    @classmethod
    def _get_binary_operator(cls, op, inv=False):
        func = dask.dataframe.core.Series._get_binary_operator(op, inv=inv)
        return cls._apply(func)

    @classmethod
    def _apply(cls, func, accessor_type=None, partial_func_name=None):
        def series_func(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, dask.dataframe.core.Series):

                def df_func(
                    df,
                    name=None,
                    func=None,
                    args=[],
                    accessor_type=None,
                    partial_func_name=None,
                    kwargs={},
                ):
                    if name is None:
                        args = [
                            df[arg] if isinstance(arg, str) else arg for arg in args
                        ]
                        return func(*args, **kwargs)
                    else:
                        if accessor_type is not None:
                            function_name = partial_func_name
                            series = getattr(df[name], accessor_type)
                        else:
                            function_name = func.__name__
                            series = df[name]

                        return getattr(series, function_name)(*args, **kwargs)

                new_kwargs = kwargs.copy()
                args = [
                    arg.name if isinstance(arg, dask.dataframe.core.Series) else arg
                    for arg in args
                ]
                new_kwargs.update(
                    {
                        "name": result.name,
                        "func": func,
                        "args": args,
                        "accessor_type": accessor_type,
                        "partial_func_name": partial_func_name,
                        "kwargs": kwargs,
                    }
                )
                ft = FunctionTransformer(
                    func=df_func, kw_args=new_kwargs, validate=False
                ).fit(result)
                result = DaskSeries(result, result.name)
                function_name = (
                    "%s-%s" % (partial_func_name, str(uuid.uuid4()))
                    if partial_func_name is not None
                    else "%s-%s" % (func.__name__, str(uuid.uuid4()))
                )
                result.transformers.append((function_name, ft))
                return result

            return result

        return series_func


# bind operators
for op in [
    operator.abs,
    operator.add,
    operator.and_,
    operator.truediv,
    operator.eq,
    operator.gt,
    operator.ge,
    operator.inv,
    operator.lt,
    operator.le,
    operator.mod,
    operator.mul,
    operator.ne,
    operator.neg,
    operator.or_,
    operator.pow,
    operator.sub,
    operator.truediv,
    operator.floordiv,
    operator.xor,
]:
    DaskSeries._bind_operator(op)


class DaskSeriesAccessor:
    """
    A DaskSeriesAccessor object.
    """

    @deprecated("2.5.2")
    def __init__(self, series_accessor, accessor_type):
        self.series_accessor = series_accessor
        self.accessor_type = accessor_type

    def __getattr__(self, item):
        attr = getattr(self.series_accessor, item)
        if (
            isinstance(attr, types.FunctionType)
            or isinstance(attr, types.MethodType)
            or isinstance(attr, partial)
        ):
            return DaskSeries._apply(
                attr, accessor_type=self.accessor_type, partial_func_name=item
            )
        else:
            return attr



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging

from oci.ai_language import AIServiceLanguageClient
from oci.data_flow import DataFlowClient
from oci.data_labeling_service import DataLabelingManagementClient
from oci.data_labeling_service_dataplane import DataLabelingClient
from oci.data_science import DataScienceClient
from oci.identity import IdentityClient
from oci.object_storage import ObjectStorageClient
from oci.secrets import SecretsClient
from oci.vault import VaultsClient

logger = logging.getLogger(__name__)


class OCIClientFactory:

    """
    A factory class to create OCI client objects. The constructor takes in config, signer and client_kwargs. `client_kwargs` is passed
    to the client constructor as key word argutments.

    Examples
    --------
    from ads.common import auth as authutil
    from ads.common import oci_client as oc

    auth = authutil.default_signer()
    oc.OCIClientFactory(**auth).object_storage # Creates Object storage client

    auth = authutil.default_signer({"timeout": 6000})
    oc.OCIClientFactory(**auth).object_storage # Creates Object storage client with timeout set to 6000

    auth = authutil.api_keys(config="/home/datascience/.oci/config", profile="TEST", {"timeout": 6000})
    oc.OCIClientFactory(**auth).object_storage # Creates Object storage client with timeout set to 6000 using API Key authentication

    auth = authutil.resource_principal({"timeout": 6000})
    oc.OCIClientFactory(**auth).object_storage # Creates Object storage client with timeout set to 6000 using resource principal authentication

    auth = authutil.create_signer("instance_principal")
    oc.OCIClientFactory(**auth).object_storage # Creates Object storage client using instance principal authentication
    """

    def __init__(self, config={}, signer=None, client_kwargs=None):
        self.config = config
        self.signer = signer
        self.client_kwargs = client_kwargs

    def _client_impl(self, client):
        client_map = {
            "object_storage": ObjectStorageClient,
            "data_science": DataScienceClient,
            "dataflow": DataFlowClient,
            "secret": SecretsClient,
            "vault": VaultsClient,
            "identity": IdentityClient,
            "ai_language": AIServiceLanguageClient,
            "data_labeling_dp": DataLabelingClient,
            "data_labeling_cp": DataLabelingManagementClient,
        }

        assert (
            client in client_map
        ), f"Invalid client name. Client name not found in {client_map.keys()}"
        return client_map[client]

    def _validate_auth_param(self, auth):
        if not isinstance(auth, dict):
            raise ValueError("auth parameter should be of type dictionary")
        if "config" in auth and not isinstance(auth["config"], dict):
            raise ValueError("auth[config] should be of type dict")
        if "signer" in auth and auth["signer"] is None and len(auth["config"]) == 0:
            raise ValueError(
                "Signer is None and auth[config] is empty. Either assign config or if you are using resource principal, set resource principal signer to signer and {} for config"
            )
        return True

    def create_client(self, client_name):
        assert (
            client_name is not None and client_name != ""
        ), "Client name cannot be empty"
        client = (
            self._client_impl(client_name)
            if isinstance(client_name, str)
            else client_name
        )
        kwargs = self.client_kwargs or dict()
        return client(config=self.config, signer=self.signer, **kwargs)

    @property
    def object_storage(self):
        return self.create_client("object_storage")

    @property
    def identity(self):
        return self.create_client("identity")

    @property
    def data_science(self):
        return self.create_client("data_science")

    @property
    def dataflow(self):
        return self.create_client("dataflow")

    @property
    def secret(self):
        return self.create_client("secret")

    @property
    def vault(self):
        return self.create_client("vault")

    @property
    def ai_language(self):
        return self.create_client("ai_language")

    @property
    def data_labeling_cp(self):
        return self.create_client("data_labeling_cp")

    @property
    def data_labeling_dp(self):
        return self.create_client("data_labeling_dp")
